from joule.models import Module, Worker, pipes
from joule.models.supervisor import Supervisor
from typing import List
import functools
import asyncio
import unittest

from tests import helpers
from tests.helpers import AsyncTestCase


class TestSupervisor(AsyncTestCase):
    NUM_WORKERS = 4

    def setUp(self):
        super().setUp()
        self.workers: List[Worker] = []
        for x in range(TestSupervisor.NUM_WORKERS):
            module = Module(name="m%d" % x, exec_cmd="", uuid=x)
            self.workers.append(Worker(module, socket_dir="/tmp", echo_module_logs=False))
        self.supervisor = Supervisor(self.workers, [], None)

    def test_returns_workers(self):
        self.assertEqual(self.workers, self.supervisor.workers)

    def test_starts_and_stops_workers(self):
        started_uuids = []
        stopped_uuids = []

        async def mock_run(uuid, _):
            started_uuids.append(uuid)

        async def mock_stop(uuid):
            stopped_uuids.append(uuid)

        for worker in self.workers:
            worker.run = functools.partial(mock_run, worker.module.uuid)
            worker.stop = functools.partial(mock_stop, worker.module.uuid)

        asyncio.run(self.supervisor.start())
        asyncio.run(self.supervisor.stop())
        # make sure each worker ran
        for worker in self.workers:
            self.assertIn(worker.module.uuid, started_uuids)
            self.assertIn(worker.module.uuid, stopped_uuids)

        self.assertEqual(len(started_uuids), TestSupervisor.NUM_WORKERS)
        self.assertEqual(len(stopped_uuids), TestSupervisor.NUM_WORKERS)

    def test_returns_sockets(self):
        for worker in self.workers:
            uuid = worker.module.uuid
            if worker.is_app:
                self.assertEqual(worker.app_socket,
                                self.supervisor.get_module_socket(uuid))
            else:
                self.assertIsNone(self.supervisor.get_module_socket(uuid))
        # returns None if the uuid doesn't exist
        bad_uuid = 68
        self.assertIsNone(self.supervisor.get_module_socket(bad_uuid))

    def test_restarts_producers(self):
        restarted = False

        async def mock_restart():
            nonlocal restarted
            restarted = True

        s = helpers.create_stream("test", "float32_3")
        self.workers[0].module.outputs['output'] = s
        self.workers[0].restart = mock_restart
        with self.assertLogs(level="WARNING") as logs:
            asyncio.run(
                self.supervisor.restart_producer(s, msg="test"))
        self.assertIn('restarting', ''.join(logs.output).lower())
        self.assertTrue(restarted)
        # check the worker logs
        self.assertIn("restarting", ''.join(self.workers[0].logs).lower())
        self.assertIn("test", ''.join(self.workers[0].logs).lower())

    def test_subscribes_to_remote_inputs(self):
        remote_stream = helpers.create_stream('remote', 'float64_2')
        remote_stream.set_remote('http://remote:3000', '/path/to/stream')
        p1 = pipes.LocalPipe(layout='float64_2', name='p1')
        p2 = pipes.LocalPipe(layout='float64_2', name='p2')
        subscription_requests = 0

        class MockNode:
            def __init__(self):
                pass

            async def data_subscribe(self, stream):
                nonlocal subscription_requests
                pipe = pipes.LocalPipe(layout='float64_2')
                subscription_requests += 1
                return pipe

            async def close(self):
                pass

        def get_mock_node(name: str):
            #TODO: name should just be the CN of the node
            self.assertEqual(name, "http://remote:3000")
            return MockNode()

        async def setup():
            self.supervisor.get_node = get_mock_node
            self.supervisor.subscribe(remote_stream, p1)
            self.supervisor.subscribe(remote_stream, p2)

            self.supervisor.task = asyncio.sleep(0)
        asyncio.run(setup())
        asyncio.run(self.supervisor.stop())

        # make sure there is only one connection to the remote
        self.assertEqual(subscription_requests, 1)
        # p2 should be subscribed to p1
        self.assertIs(p1.subscribers[0], p2)

        # now "stop" the supervisor and the pipes should be closed

        self.assertTrue(p1.closed)
        self.assertTrue(p2.closed)

    @unittest.skip("Test is broken, but this hasn't been used so ignore for now")
    def test_subscribes_to_remote_outputs(self):
        remote_stream = helpers.create_stream('remote', 'float64_2')
        remote_stream.set_remote('remote:3000', '/path/to/stream')
        module = Module(name="remote_output", exec_cmd="", uuid=0)
        module.outputs = {'output': remote_stream}
        worker = Worker(module, socket_dir="/tmp", echo_module_logs=False)

        async def mock_run(subscribe):
            await asyncio.sleep(0.1)

        worker.run = mock_run

        output_requested = False

        class MockNode:
            
            async def data_write(self, stream):
                nonlocal output_requested
                output_requested = True
                pipe = pipes.LocalPipe(layout='float64_2')
                return pipe

            async def close(self):
                # API compatability
                pass

        def get_mock_node(name: str):
            #TODO: name should just be the CN of the node
            self.assertEqual(name, "remote:3000")
            return MockNode()

        supervisor = Supervisor([worker], [], get_mock_node)

        asyncio.run(supervisor.start())
        self.supervisor.task = asyncio.sleep(0.1)
        asyncio.run(self.supervisor.start())
        asyncio.run(self.supervisor.stop())

        # pipe should be created
        self.assertTrue(output_requested)
        # worker should be producing data into the pipe
        self.assertEqual(1, len(worker.subscribers[remote_stream]))
        asyncio.run(supervisor.stop())



import testing.postgresql
from aiohttp.test_utils import AioHTTPTestCase
from sqlalchemy.orm import Session
from aiohttp import web
import aiohttp
import asyncio
import datetime

from joule.models import DataStream, Element
import joule.controllers
from tests.controllers.helpers import create_db, MockStore
from joule import app_keys
from joule.constants import EndPoints

psql_key = web.AppKey("psql", testing.postgresql.Postgresql)
class TestStreamControllerErrors(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        self.app[psql_key].stop()
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0
        app[app_keys.db], app[psql_key] = create_db(["/folder1/stream1:float32[x, y, z]",
                                                     "/folder1/same_name:float32[x, y, z]",
                                                     "/folder2/deeper/stream2:int16[val1, val2]",
                                                     "/folder_x1/same_name:float32[x, y, z]",
                                                     "/folder_x2/same_name:int16[val1, val2]"
                                                     ])
        app[app_keys.data_store] = MockStore()
        return app

    async def test_stream_info(self):
        # must specify an id or a path
        resp: aiohttp.ClientResponse = await self.client.request("GET", EndPoints.stream)
        self.assertEqual(resp.status, 400)
        # return "not found" on bad id
        resp: aiohttp.ClientResponse = await self.client.request("GET", f"{EndPoints.stream}?id=2348")
        self.assertEqual(resp.status, 404)
        # return "not found" on bad path
        payload = {'path': "/bad/path"}
        resp = await self.client.request("GET", EndPoints.stream, params=payload)
        self.assertEqual(resp.status, 404)

    async def test_stream_move(self):
        db: Session = self.app[app_keys.db]

        # must be json
        resp = await self.client.put(EndPoints.stream_move, data={"dest_path": "/new/folder3"})
        self.assertEqual(resp.status, 400)
        self.assertIn("json", await resp.text())
        # must specify an id or a path
        resp = await self.client.put(EndPoints.stream_move, json={"dest_path": "/new/folder3"})
        self.assertEqual(resp.status, 400)
        # must specify a destination
        resp = await self.client.put(EndPoints.stream_move, json={"src_path": "/folder1/stream1"})
        self.assertEqual(resp.status, 400)
        # return "not found" on bad id
        resp = await self.client.put(EndPoints.stream_move, json={"src_id": 2348,
                                                                "dest_path": "/x/y"})
        self.assertEqual(resp.status, 404)
        # return "not found" on bad path
        resp = await self.client.put(EndPoints.stream_move, json={"src_path": "/folder2/deeper",
                                                                "dest_path": "/x/y"})
        self.assertEqual(resp.status, 404)
        # error on bad destination
        resp = await self.client.put(EndPoints.stream_move, json={"src_path": "/folder2/deeper",
                                                                "dest_path": "malformed"})
        self.assertEqual(resp.status, 404)
        resp = await self.client.put(EndPoints.stream_move, json={"src_path": "/folder1/stream1",
                                                                "dest_path": "invalid"})
        self.assertEqual(resp.status, 400)

        # cannot conflict with an existing stream in the destination
        resp = await self.client.put(EndPoints.stream_move, json={"src_path": "/folder_x1/same_name",
                                                                "dest_path": "/folder_x2"})
        self.assertEqual(resp.status, 400)
        # cannot move streams with a fixed configuration
        my_stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        my_stream.is_configured = True
        resp = await self.client.put(EndPoints.stream_move, json={"src_path": "/folder1/stream1",
                                                                "dest_path": "/folder8"})
        self.assertEqual(resp.status, 400)
        self.assertIn('locked', await resp.text())

    async def test_stream_create(self):
        # must be json
        resp = await self.client.post(EndPoints.stream, data={"bad_values": "not_json"})
        self.assertEqual(resp.status, 400)
        self.assertIn("json", await resp.text())

        # must specify a stream
        resp = await self.client.post(EndPoints.stream, json={"dest_path": "/folder2/deeper"})
        self.assertEqual(resp.status, 400)
        # invalid dest_path
        resp = await self.client.post(EndPoints.stream, json={"dest_path": "notapath"})
        self.assertEqual(resp.status, 400)
        # must specify a path
        new_stream = DataStream(name="test", datatype=DataStream.DATATYPE.FLOAT32,
                                updated_at=datetime.datetime.utcnow())
        new_stream.elements = [Element(name="e%d" % j, index=j,
                                       display_type=Element.DISPLAYTYPE.CONTINUOUS) for j in range(3)]
        resp = await self.client.post(EndPoints.stream, json={"stream": new_stream.to_json()})
        self.assertEqual(resp.status, 400)
        # element names must be unique
        e1_name = new_stream.elements[1].name
        new_stream.elements[1].name = new_stream.elements[2].name
        resp = await self.client.post(EndPoints.stream, json={
            "dest_path": "/folder2/deeper",
            "stream": new_stream.to_json()})
        self.assertEqual(resp.status, 400)
        self.assertIn("names must be unique", await resp.text())
        new_stream.elements[1].name = e1_name  # restore the original name

        # stream must have a unique name in the folder
        new_stream.name = "stream1"
        resp = await self.client.post(EndPoints.stream, json={"stream": new_stream.to_json(),
                                                            "dest_path": "/folder1"})
        self.assertEqual(resp.status, 400)
        self.assertIn("same name", await resp.text())
        # invalid dest_path
        resp = await self.client.post(EndPoints.stream, json={"stream": new_stream.to_json(),
                                                            "dest_path": "notapath"})
        self.assertEqual(resp.status, 400)
        # stream must have a name
        new_stream.name = ""
        resp = await self.client.post(EndPoints.stream, json={"stream": new_stream.to_json(),
                                                            "dest_path": "/a/valid/path"})
        self.assertEqual(resp.status, 400)
        self.assertIn("name", await resp.text())
        new_stream.name = "test"
        # stream must have at least one element
        new_stream.elements = []
        resp = await self.client.post(EndPoints.stream, json={"stream": new_stream.to_json(),
                                                            "dest_path": "/a/valid/path"})
        self.assertEqual(resp.status, 400)
        self.assertIn("element", await resp.text())

        # invalid stream json (test all different exception paths)
        resp = await self.client.post(EndPoints.stream, json={"dest_path": "/path/to",
                                                            "stream": 'notjson'})
        self.assertEqual(resp.status, 400)
        self.assertIn("JSON", await resp.text())

        json = new_stream.to_json()
        json["datatype"] = "invalid"
        resp = await self.client.post(EndPoints.stream, json={"dest_path": "/path/to",
                                                            "stream": json})
        self.assertEqual(resp.status, 400)
        self.assertIn("specification", await resp.text())
        self.assertIn("datatype", await resp.text())

        del json["datatype"]
        resp = await self.client.post(EndPoints.stream, json={"dest_path": "/path/to",
                                                            "stream": json})
        self.assertEqual(resp.status, 400)
        self.assertIn("datatype", await resp.text())

        # incorrect stream format
        resp = await self.client.post(EndPoints.stream, json={"path": "/path/to",
                                                            "stream": '{"invalid": 2}'})
        self.assertEqual(resp.status, 400)

    async def test_stream_delete(self):
        db: Session = self.app[app_keys.db]

        resp = await self.client.delete(EndPoints.stream, params={})
        self.assertEqual(resp.status, 400)
        # return "not found" on bad id
        resp = await self.client.delete(EndPoints.stream, params={"id": 2348})
        self.assertEqual(resp.status, 404)
        # return "not found" on bad path
        resp = await self.client.delete(EndPoints.stream, params={"path": "/bad/path"})
        self.assertEqual(resp.status, 404)
        # cannot delete locked streams
        my_stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        my_stream.is_configured = True
        resp = await self.client.delete(EndPoints.stream, params={"id": my_stream.id})
        self.assertEqual(resp.status, 400)
        self.assertIn('locked', await resp.text())
        self.assertIsNotNone(db.query(DataStream).filter_by(name="stream1").one())

    async def test_stream_update(self):
        db: Session = self.app[app_keys.db]
        my_stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()

        # must be json
        resp = await self.client.put(EndPoints.stream, data={"bad_values": "not_json"})
        self.assertEqual(resp.status, 400)
        self.assertIn("json", await resp.text())

        # invalid element display_type, nothing should be saved
        payload = {
            "id": my_stream.id,
            "stream":
                {"name": "new name",
                 "elements": [{"display_type": "invalid", "name": "new"},
                              {"name": "new1"},
                              {"name": "new3"}]}
        }
        resp = await self.client.put(EndPoints.stream, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertIn('display_type', await resp.text())

        my_stream: DataStream = db.get(DataStream, my_stream.id)
        self.assertEqual("stream1", my_stream.name)
        elem_0 = [e for e in my_stream.elements if e.index == 0][0]
        self.assertEqual(elem_0.display_type, Element.DISPLAYTYPE.CONTINUOUS)
        self.assertEqual(elem_0.name, 'x')

        # request must specify an id
        payload = {
            "stream": {"name": "new name"}
        }
        resp = await self.client.put(EndPoints.stream, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertIn('id', await resp.text())

        # request must have streams attr
        payload = {
            "id": my_stream.id
        }
        resp = await self.client.put(EndPoints.stream, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertIn('stream', await resp.text())

        # streams attr must be valid json
        payload = {
            "id": my_stream.id,
            "stream": "notjson"
        }
        resp = await self.client.put(EndPoints.stream, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertIn('JSON', await resp.text())

        # cannot modify locked streams
        my_stream.is_configured = True
        payload = {
            "id": my_stream.id,
            "stream": {"name": "new name"}
        }
        resp = await self.client.put(EndPoints.stream, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertIn('locked', await resp.text())

        # stream must exist
        payload = {
            "id": 4898,
            "stream": {"name": "new name"}
        }
        resp = await self.client.put(EndPoints.stream, json=payload)
        self.assertEqual(resp.status, 404)
        self.assertIn('exist', await resp.text())

        # stream name must be unique in the folder
        my_stream.is_configured = False  # unlock
        payload = {
            "id": my_stream.id,
            "stream": {"name": "same_name"}
        }
        resp = await self.client.put(EndPoints.stream, json=payload)
        self.assertEqual(resp.status, 400)
        self.assertIn('same name', await resp.text())



# Dataclass for API endpoints in Joule
from dataclasses import dataclass

@dataclass
class EndPoints:
    root = '/'

    db_connection = '/db/connection.json'
    db_info = '/dbinfo'
    
    version = '/version'
    version_json  = '/version.json'

    event = '/event.json'
    event_move = '/event/move.json'
    event_data = '/event/data.json'
    event_data_count = '/event/data/count.json'
    
    stream = '/stream.json'
    stream_move = '/stream/move.json'
    stream_annotations = '/stream/annotations.json'
    
    folder = '/folder.json'
    folders = '/folders.json'
    folder_move = '/folder/move.json'
    
    data = '/data'
    data_json = '/data.json'
    data_intervals = '/data/intervals.json'
    data_decimate = '/data/decimate.json'
    data_consolidate = '/data/consolidate.json'
    
    module = '/module.json'
    modules = '/modules.json'
    module_logs = '/module/logs.json'
    
    app_auth = '/app/auth'
    app_json = '/app.json'
    
    proxy = '/proxy.json'
    proxies = '/proxies.json'
    
    master = '/master.json'
    masters = '/masters.json'
    
    follower = '/follower.json'
    followers = '/followers.json'

    annotation = '/annotation.json'
    annotations = '/annotations.json'
    annotations_info = '/annotations/info.json'

@dataclass
class ApiErrorMessages:
    stream_does_not_exist = 'stream does not exist'
    folder_does_not_exist = 'folder does not exist'
    specify_id_or_path = 'specify an id or a path'
    start_must_be_before_end = '[start] must be before [end]'
    start_and_end_must_be_integers = '[start] and [end] must be integers'
    invalid_filter_parameter = 'invalid filter parameter'
    id_not_found = 'id not found'
    folder_is_locked = 'folder is locked'
    # Error messages that require formatting
    f_parameter_must_be_an_int = "parameter {parameter} must be an int"

@dataclass
class ConfigFiles:
    default_node='default_node.txt'
    nodes = 'nodes.json'
    main_config = '/etc/joule/main.conf'




#!/usr/bin/env python3

import joule.client
import numpy as np


class MergeSum(joule.client.FilterModule):
    " Sums two streams "

    def custom_args(self, parser):
        pass
        
    async def run(self, parsed_args, inputs, outputs):
        in1 = inputs["input1"]
        in2 = inputs["input2"]
        stream_out = outputs["output"]
        await self.align_inputs(in1, in2)
        
        while(1):
            sarray1 = await in1.read(flatten=True)
            sarray2 = await in2.read(flatten=True)
            size = min(len(sarray1), len(sarray2))
            
            sarray1[:size, 1:] += sarray2[:size, 1:]
            await stream_out.write(sarray1[:size])
            in1.consume(size)
            in2.consume(size)

    async def align_inputs(self, in1, in2):
        while(1):
            # find the max starting time
            sarray1 = await in1.read()
            sarray2 = await in2.read()
            max_ts = max(sarray1['time'][0],
                         sarray2['time'][0])
            
            idx1 = np.argwhere(sarray1['time'] == max_ts)
            if(len(idx1) == 0):
                in1.consume(len(sarray1))
                continue

            idx2 = np.argwhere(sarray2['time'] == max_ts)
            if(len(idx2) == 0):
                in2.consume(len(sarray2))
                continue
            
            # found valid indices in both arrays
            idx1 = idx1[0, 0]
            idx2 = idx2[0, 0]
            
            in1.consume(idx1)
            in2.consume(idx2)
            return

        
if __name__ == "__main__":
    r = MergeSum()
    r.start()



import warnings
from aiohttp.test_utils import unused_port
from click.testing import CliRunner
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

warnings.simplefilter('always')


class TestFolderMove(FakeJouleTestCase):

    def test_folder_move(self):
        server = FakeJoule()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['folder', 'move', '/folder/src', '/folder/dest'])
        self.assertEqual(result.exit_code, 0)
        (path, destination) = self.msgs.get()
        self.assertEqual(path, '/folder/src')
        self.stop_server()

    def test_when_folder_does_not_exist(self):
        server = FakeJoule()
        server.response = "folder does not exist"
        server.http_code = 404
        server.stub_folder_move = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['folder', 'move', '/bad/path', '/folder/dest'])
        self.assertIn("Error", result.output)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        server.response = "test error"
        server.http_code = 500
        server.stub_folder_move = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['folder', 'move', '/folder/src', '/folder/dest'])
        self.assertIn('500', result.output)
        self.assertIn("test error", result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
from sqlalchemy.orm import Session
import asyncio
from unittest.mock import patch
from joule.models import master, Master
import joule.controllers
from tests.controllers.helpers import create_db, MockStore
from joule import app_keys
from joule.constants import EndPoints

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestMasterControllerErrors(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        await asyncio.to_thread(self.app[psql_key].stop)
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0

        app[app_keys.db], app[psql_key] = await asyncio.to_thread(lambda: create_db([]))
        app[app_keys.data_store] = MockStore()
        app[app_keys.name] = "test"
        app[app_keys.port] = 443
        app[app_keys.scheme] = "http"
        app[app_keys.base_uri] = "/"
        app[app_keys.cafile] = "stubbed_cafile"

        # create a master user to grant access
        self.grantor = Master(name="grantor", key="grantor_key",
                              type=Master.TYPE.USER)
        app[app_keys.db].add(self.grantor)
        app[app_keys.db].commit()
        return app


    async def test_adds_master_user(self):
        # must specify a valid master type
        payload = {
            "master_type": "invalid",
            "identifier": "johndoe"
        }
        resp = await self.client.post(EndPoints.master, json=payload,
                                      headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 400)
        self.assertIn('master_type', await resp.text())

        # master_type must be in request
        payload = {
            "identifier": "johndoe"
        }
        resp = await self.client.post(EndPoints.master, json=payload,
                                      headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 400)
        self.assertIn('master_type', await resp.text())

        # identifier must be in request
        payload = {
            "master_type": "joule"
        }
        resp = await self.client.post(EndPoints.master, json=payload,
                                      headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 400)
        self.assertIn('identifier', await resp.text())

        # if an API key is specified it must be long enough
        payload = {
            "master_type": "user",
            "identifier": "johndoe",
            "api_key": "short"
        }
        resp = await self.client.post(EndPoints.master, json=payload,
                                      headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 400)
        self.assertIn('API Key', await resp.text())

        # if the identifier is not a URL and cannot be turned into one
        # the request should fail
        with patch('joule.controllers.master_controller.detect_url') as mock_detect_url:
            mock_detect_url.return_value = None
            # For joule nodes...
            payload = {
                "master_type": "joule",
                "identifier": "invalid"
            }
            resp = await self.client.post(EndPoints.master, json=payload,
                                          headers={'X-API-KEY': "grantor_key"})
            self.assertEqual(resp.status, 400)
            self.assertIn('cannot connect', await resp.text())
            # For lumen nodes...
            payload = {
                "master_type": "lumen",
                "identifier": "invalid",
                "lumen_params": {"key": "stubbed"}
            }
            resp = await self.client.post(EndPoints.master, json=payload,
                                          headers={'X-API-KEY': "grantor_key"})
            self.assertEqual(resp.status, 400)
            self.assertIn('cannot connect', await resp.text())

    async def test_master_controller_delete_errors(self):
        # cannot delete yourself
        payload = {
            "master_type": "user",
            "name": "grantor"
        }
        resp = await self.client.delete(EndPoints.master, params=payload,
                                        headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 400)
        self.assertIn('cannot delete yourself', await resp.text())

        # must specify a name and master_type
        payload = {
            "name": "johndoe"
        }
        resp = await self.client.delete(EndPoints.master, params=payload,
                                        headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 400)
        self.assertIn('master_type', await resp.text())

        # master must exist
        payload = {
            "master_type": "user",
            "name": "not-present"
        }
        resp = await self.client.delete(EndPoints.master, params=payload,
                                        headers={'X-API-KEY': "grantor_key"})
        self.assertEqual(resp.status, 404)
        self.assertIn('not a master', await resp.text())


import unittest
from joule.errors import ConfigurationError
from joule.models import module  # for helper functions
from tests import helpers


class TestModuleErrors(unittest.TestCase):

    def setUp(self):
        self.base_config = helpers.parse_configs(
            """[Main]
                 name = test
                 description = test_description                 
                 path = /some/path/to/data
                 datatype = float32
                 keep = 1w
                 decimate = yes
               [Element1]
                 name = e1
            """)

    def test_errors_on_missing_main(self):
        self.base_config.remove_section("Main")
        with self.assertRaisesRegex(ConfigurationError, 'Main'):
            module.from_config(self.base_config)

    def test_errors_on_bad_name(self):
        self.base_config.remove_option('Main', 'name')
        with self.assertRaisesRegex(ConfigurationError, 'name'):
            module.from_config(self.base_config)
        self.base_config.set('Main', 'name', '')
        with self.assertRaisesRegex(ConfigurationError, 'name'):
            module.from_config(self.base_config)




from aiohttp.test_utils import AioHTTPTestCase
from aiohttp import web
import aiohttp
import json
import asyncio
import joule.controllers
from joule.models import Folder, EventStream
from joule.models.folder import find as find_folder
from tests.controllers.helpers import create_db, MockSupervisor, MockEventStore
from joule.api.event_stream import EventStream as ApiEventStream, from_json as event_stream_from_json
from joule.api.event import Event
from joule import app_keys
from joule.constants import EndPoints

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestEventController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        await asyncio.to_thread(self.app[psql_key].stop)
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        stream1 = ApiEventStream(name="test",description="test stream",keep_us=10e6, 
                              event_fields={"field1": "string",
                                           "field2": 'category:["1","2","3"]'})
    
        app[app_keys.db], app[psql_key] = await asyncio.to_thread(lambda: create_db(["/other/folder/stub_stream:float32[x,y,z]"],[stream1]))
        self.test_stream_id = app[app_keys.db].query(EventStream).filter_by(name="test").one().id
        self.test_stream = stream1 # stream2 is to test for name collisions
        self.test_stream.id = self.test_stream_id # set the id to match the database
        app[app_keys.event_store] = MockEventStore()
        self.supervisor = MockSupervisor()
        app[app_keys.supervisor] = self.supervisor
        return app
    
    async def test_update_event_stream(self):
        db = self.app[app_keys.db]
        stream = db.query(EventStream).filter_by(id=self.test_stream_id).one()
        prev_update = stream.updated_at
        self.test_stream.name = "new_name"
        self.test_stream.description = "new description"
        self.test_stream.keep_us = 20e6
        self.test_stream.event_fields = {"new_field": "string"}
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event, json={"id": self.test_stream_id, "stream": self.test_stream.to_json()})
        # print the response text if the test fails
        if resp.status != 200:
            print(await resp.text())
        self.assertEqual(resp.status, 200)
        # returns the stream as JSON
        data = await resp.json()
        stream = event_stream_from_json(data)
        self.assertEqual(stream.name, "new_name")
        # database is updated
        stream = db.query(EventStream).filter_by(id=self.test_stream_id).one()
        self.assertEqual(stream.name, "new_name")
        self.assertEqual(stream.description, "new description")
        self.assertEqual(stream.keep_us, 20e6)
        self.assertEqual(stream.event_fields, {"new_field": "string"})
        # updated_at is more recent 
        self.assertGreater(stream.updated_at, prev_update)

    async def test_create_event_streams(self):
        store: MockEventStore = self.app[app_keys.event_store]
        new_stream = ApiEventStream(name="test",description="test stream",keep_us=10e6, 
                                 event_fields={"field1": "string",
                                               "field2": 'category:["1","2","3"]'})
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event, json={"dest_path": "/folder1", "stream": new_stream.to_json()})
        self.assertEqual(resp.status, 200)
        self.assertTrue(store.create_called)
        self.assertEqual(store.new_stream.folder.name,'folder1')
        self.assertEqual(store.new_stream.name,'test')
        self.assertEqual(store.new_stream.description,'test stream')
        self.assertEqual(store.new_stream.keep_us,10e6)
        self.assertEqual(store.new_stream.event_fields['field2'], 'category:["1","2","3"]')

        # can also create a stream with a destination id
        dest_id = self.app[app_keys.db].query(Folder).filter_by(name="folder1").one().id
        new_stream.name="test2"
        store.create_called = False
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event, json={"dest_id": dest_id, "stream": new_stream.to_json()})
        self.assertEqual(resp.status, 200)
        self.assertTrue(store.create_called)
        self.assertEqual(store.new_stream.folder.name,'folder1')
        self.assertEqual(store.new_stream.name,'test2')
    
        # names must be unique within a folder
        store.create_called = False
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event, json={"dest_id": dest_id, "stream": new_stream.to_json()})
        self.assertEqual(resp.status, 400)
        self.assertFalse(store.create_called)
        msg = await resp.text()
        self.assertIn('same name', msg)

    async def test_destroy_event_stream_by_path(self):
        self.assertEqual(self.app[app_keys.db].query(EventStream).filter_by(id=self.test_stream_id).count(), 1)
        prev_update = self.app[app_keys.db].query(Folder).filter_by(name="events").one().updated_at
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.event, params={"path": "/events/test"})
        self.assertEqual(resp.status, 200)
        self.assertEqual(self.app[app_keys.db].query(EventStream).filter_by(id=self.test_stream_id).count(), 0)
        # make sure the parent folder timestamp is updated
        updated_at = self.app[app_keys.db].query(Folder).filter_by(name="events").one().updated_at
        self.assertGreater(updated_at, prev_update)
    
    async def test_destroy_event_stream_by_id(self):
        self.assertEqual(self.app[app_keys.db].query(EventStream).filter_by(id=self.test_stream_id).count(), 1)
        prev_update = self.app[app_keys.db].query(Folder).filter_by(name="events").one().updated_at
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.event, params={"id": self.test_stream_id})
        self.assertEqual(resp.status, 200)
        self.assertEqual(self.app[app_keys.db].query(EventStream).filter_by(id=self.test_stream_id).count(), 0)
        # make sure the parent folder timestamp is updated
        updated_at = self.app[app_keys.db].query(Folder).filter_by(name="events").one().updated_at
        self.assertGreater(updated_at, prev_update)
    
    async def test_get_event_stream(self):
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event, params={"id": 1})
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        stream = event_stream_from_json(data)
        self.assertEqual(stream.name, "test")

        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.event, params={"path": "/events/test"})
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        stream = event_stream_from_json(data)
        self.assertEqual(stream.name, "test")

    async def test_move_event_stream(self):
        orig_parent = find_folder("/events", self.app[app_keys.db])
        expected_new_parent = find_folder("/other/folder", self.app[app_keys.db])
        new_parent_orig_ts = expected_new_parent.updated_at
        old_parent_orig_ts = find_folder("/events", self.app[app_keys.db]).updated_at

        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event_move, json={"dest_path": "/other/folder", "src_path": "/events/test"})
        self.assertEqual(resp.status, 200)
        event_stream = event_stream_from_json((await resp.json())['stream'])
        # returns the stream
        self.assertEqual(event_stream.name, "test")
        # stream is moved, make sure the event stream's folder id matches the new parent
        actual_parent_id = self.app[app_keys.db].query(EventStream).filter_by(name="test").one().folder_id
        self.assertEqual(actual_parent_id, expected_new_parent.id)
        # make sure the parent folder timestamps are updated
        new_parent_updated_ts = find_folder("/other/folder", self.app[app_keys.db]).updated_at
        self.assertGreater(new_parent_updated_ts, new_parent_orig_ts)
        old_parent_updated_ts = find_folder("/events", self.app[app_keys.db]).updated_at
        self.assertGreater(old_parent_updated_ts, old_parent_orig_ts)

        # also works with ids, move it back to the original location
        resp: aiohttp.ClientResponse = await \
            self.client.put(EndPoints.event_move, json={"dest_id": orig_parent.id, 
                                                        "src_id": self.test_stream_id})
        self.assertEqual(resp.status, 200)
        actual_parent_id = self.app[app_keys.db].query(EventStream).filter_by(name="test").one().folder_id
        self.assertEqual(actual_parent_id, orig_parent.id)

    async def test_write_events(self):
        event_store = self.app[app_keys.event_store]
        ### writes events given an event stream id
        events = [Event(start_time=i,end_time=i+1000,content={"field1": i}, 
                        event_stream_id=self.test_stream.id, id=i, node_uuid='test') for i in range(0,10000,1000)]
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event_data, json={"id": self.test_stream_id, "events": [e.to_json(destination_node_uuid='test') for e in events]})
        self.assertEqual(resp.status, 200)
        # make sure the events are written to the event store
        self.assertTrue(event_store.upsert_called)
        self.assertEqual(event_store.upserted_events, [e.to_json(destination_node_uuid='test') for e in events])
        event_store.reset()

        ### writes events given an event stream path
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event_data, json={"path": "/events/test", "events": [e.to_json(destination_node_uuid='test') for e in events]})
        self.assertEqual(resp.status, 200)
        # make sure the events are written to the event store
        self.assertTrue(event_store.upsert_called)
        self.assertEqual(event_store.upserted_events, [e.to_json(destination_node_uuid='test') for e in events])
        # the correct event_stream_id is used
        self.assertEqual(event_store.upserted_events[0]['event_stream_id'], self.test_stream_id)
        event_store.reset()

        ### resets ID field when events do not have the right event_stream_id
        events[0].event_stream_id = 999
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.event_data, json={"id": self.test_stream_id, "events": [e.to_json(destination_node_uuid='test') for e in events]})
        self.assertEqual(resp.status, 200)
        # make sure the events are written to the event store
        self.assertTrue(event_store.upsert_called)
        modified_event = event_store.upserted_events[0]
        self.assertIsNone(modified_event['id']) # ID is removed so it will be treated as a new event
        self.assertEqual(modified_event['event_stream_id'], self.test_stream_id) # event_stream_id is corrected
        # the rest of the events are unchanged
        self.assertEqual(event_store.upserted_events[1:], [e.to_json(destination_node_uuid='test') for e in events[1:]])

    async def test_event_stream_count_events(self):
        event_store = self.app[app_keys.event_store]

        ### Works with default parameters
        resp = await self.client.get(EndPoints.event_data_count, 
                                     params={"id": self.test_stream_id})
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data['count'], 0) # no events loaded to will always return 0
        # make sure the event store is queried
        self.assertTrue(event_store.count_called)
        # default parameters passed to count
        target_stream, start, end, json_filter, include_on_going_events = event_store.count_params
        self.assertTupleEqual((start, end, json_filter, include_on_going_events), (None, None, None, False))
        self.assertEqual(target_stream.id, self.test_stream_id)
        event_store.reset()

        ### Can specify optional parameters
        json_filter = '[[["id","=","3"]]]'
        start = 0
        end = 100
        include_on_going_events = True
        resp = await self.client.get(EndPoints.event_data_count, 
                                     params={"id": self.test_stream_id,
                                             "start": start,
                                             "end": end,
                                             "filter": json.dumps(json_filter),
                                             "include_on_going_events": 0})
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data['count'], 0)
        self.assertTrue(event_store.count_called)
        # optional parameters are correctly parsed
        self.assertEqual(event_store.count_params[0].id, self.test_stream.id)
        self.assertEqual(event_store.count_params[1], start)
        self.assertEqual(event_store.count_params[2], end)
        self.assertEqual(event_store.count_params[3], json_filter)
        self.assertFalse(event_store.count_params[4])
        event_store.reset()

        ### Can find by path
        resp = await self.client.get(EndPoints.event_data_count, 
                                     params={"path": "/events/test"})
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data['count'], 0)
        self.assertTrue(event_store.count_called)
        self.assertEqual(event_store.count_params[0].id, self.test_stream.id)

    async def test_event_stream_read_events(self):
        event_store = self.app[app_keys.event_store]
        mock_events = [{"start_time":i,"end_time":i+1000,"content": {"field1": i}, 
                        "event_stream_id":self.test_stream_id, "id":i} for i in range(0,10000,1000)]
        event_store.set_events(mock_events)

        ### Works with default parameters
        resp = await self.client.get(EndPoints.event_data, 
                                     params={"id": self.test_stream_id, "limit": 20})
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertTrue(event_store.count_called)
        self.assertEqual(len(data['events']), 10)
        event_store.reset()

        ### Returns histogram when there are too many events
        event_store.set_events(mock_events)
        resp = await self.client.get(EndPoints.event_data, 
                                     params={"id": self.test_stream_id, "limit": len(mock_events)-1})
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data['type'],'histogram')
        # mock histogram function returns an empty []
        self.assertListEqual(data['events'], [])
        self.assertTrue(event_store.count_called)
        self.assertFalse(event_store.extract_called)
        self.assertTrue(event_store.histogram_called)
        event_store.reset()

        ### Can specify path instead of id and optional parameters, soft limit
        event_store.set_events(mock_events)
        start = 0
        end = 100
        limit = 5
        json_filter = [[["id","=","3"]]]
        resp = await self.client.get(EndPoints.event_data, 
                                     params={"path": "/events/test",
                                             "start": start,
                                             "end": end,
                                             "filter": json.dumps(json_filter),
                                             "limit": limit,
                                             'return-subset': 1,
                                             "include-ongoing-events": 1})
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertTrue(event_store.extract_called)
        self.assertFalse(event_store.histogram_called)
        # make sure parameters are parsed
        self.assertEqual(event_store.extract_params[0].id, self.test_stream.id)
        self.assertEqual(event_store.extract_params[1], start)
        self.assertEqual(event_store.extract_params[2], end)
        self.assertEqual(event_store.extract_params[3], json_filter)
        self.assertEqual(event_store.extract_params[4], limit)
        self.assertTrue(event_store.extract_params[5])
        # returns the subset of events, not a histogram because return-subset is set
        self.assertNotEqual(len(data['events']), len(mock_events))
        self.assertEqual(len(data['events']), limit)


    async def test_event_stream_remove_events(self):
        event_store = self.app[app_keys.event_store]
        
        ### Works with default parameters
        resp = await self.client.delete(EndPoints.event_data, 
                                        params={"id": self.test_stream_id})
        self.assertEqual(resp.status, 200)
        self.assertTrue(event_store.remove_called)
        event_store.reset()

        ### Can specify path instead of id and optional parameters
        start = 0
        end = 100
        json_filter = [[["id","=","3"]]]
        resp = await self.client.delete(EndPoints.event_data, 
                                        params={"path": "/events/test",
                                                "start": start,
                                                "end": end,
                                                "filter": json.dumps(json_filter)})
        self.assertEqual(resp.status, 200)
        self.assertTrue(event_store.remove_called)
        # make sure parameters are parsed
        self.assertEqual(event_store.remove_params[0].id, self.test_stream.id)
        self.assertEqual(event_store.remove_params[1], start)
        self.assertEqual(event_store.remove_params[2], end)
        self.assertEqual(event_store.remove_params[3], json_filter)


from sqlalchemy.orm import relationship, Mapped
from sqlalchemy import (Column, Integer, String, DateTime, ForeignKey)
from sqlalchemy.dialects.postgresql import BIGINT, JSONB
from typing import TYPE_CHECKING, Dict, Optional
from datetime import datetime, timezone
from sqlalchemy_utils import force_instant_defaults

from joule.errors import ConfigurationError
from joule.models.meta import Base
from joule.utilities.validators import validate_event_fields

if TYPE_CHECKING:
    from joule.models import (Folder)  
    from joule.models.data_store.event_store import StreamInfo

force_instant_defaults()


class EventStream(Base):
    """
    Attributes:
        name (str): stream name
        description (str): stream description
        decimate (bool): whether to store decimated data for event visualization
        folder (joule.Folder): parent Folder
        chunk_duration_us (int): chunk duration in microseconds, 0 for normal table
        keep_us (int): microseconds of data to keep (KEEP_NONE=0, KEEP_ALL=-1).
    """
    __tablename__ = 'eventstream'
    __table_args__ = {"schema": "metadata"}

    id: int = Column(Integer, primary_key=True)
    name: str = Column(String, nullable=False)
    event_fields: Dict[str, str] = Column(JSONB)
    chunk_duration_us: int = Column(BIGINT, default=0)

    KEEP_ALL = -1
    KEEP_NONE = 0
    keep_us: int = Column(BIGINT, default=KEEP_ALL)

    description: str = Column(String)
    folder_id: int = Column(Integer, ForeignKey('metadata.folder.id'), nullable=False)
    folder: Mapped["Folder"] = relationship("Folder", back_populates="event_streams")
    updated_at: datetime = Column(DateTime(timezone=True), nullable=False)

    def update_attributes(self, attrs: Dict) -> None:
        updated = False
        if 'name' in attrs:
            self.name = validate_name(attrs['name'])
            updated = True
        if 'description' in attrs:
            self.description = attrs['description']
            updated = True
        if 'event_fields' in attrs:
            self.event_fields = validate_event_fields(attrs['event_fields'])
            updated = True
        if 'chunk_duration_us' in attrs:
            self.chunk_duration_us = validate_chunk_duration(attrs['chunk_duration_us'])
            updated = True
        if 'keep_us' in attrs:
            self.keep_us = validate_keep_us(attrs['keep_us'])
            updated = True
        if updated:
            self.touch()

    # update timestamps on all folders in hierarchy
    def touch(self, now: Optional[datetime] = None) -> None:
        if now is None:
            now = datetime.now(timezone.utc)
        self.updated_at = now
        if self.folder is not None:
            self.folder.touch(now)

    def to_json(self, info: Dict[int, 'StreamInfo'] = None) -> Dict:
        """
          Args:
            info: optional content added to ``data_info`` field

        Returns: Dictionary of DataStream attributes

        """

        resp = {
            'id': self.id,
            'name': self.name,
            'description': self.description,
            'event_fields': self.event_fields,
            'chunk_duration_us': self.chunk_duration_us,
            'keep_us': self.keep_us,
            'updated_at': self.updated_at.isoformat()
        }

        if info is not None and self.id in info:
            resp['data_info'] = info[self.id].to_json()
        return resp

    def __str__(self):
        return "EventStream [{name}]".format(name=self.name)

    def __repr__(self):
        return "<EventStream(id=%r, name='%s')>" % (self.id, self.name)


def from_json(data: Dict) -> EventStream:
    """
    Construct an EventStream from a dictionary of attributes produced by
    :meth:`EventStream.to_json`

    Args:
        data: attribute dictionary

    Returns: EventStream

    """

    return EventStream(id=data["id"],
                       name=validate_name(data["name"]),
                       description=data["description"],
                       event_fields=validate_event_fields(data["event_fields"]),
                       chunk_duration_us=validate_chunk_duration(data["chunk_duration_us"]),
                       keep_us=validate_keep_us(data["keep_us"]),
                       updated_at=datetime.now(timezone.utc))


def validate_chunk_duration(duration):
    try:
        duration = int(duration)
    except (ValueError, TypeError):
        raise ConfigurationError("invalid chunk_duration_us, must be an integer")
    if duration < 0:
        raise ConfigurationError("invalid chunk_duration_us, must be > 0")
    return duration


def validate_keep_us(keep_us):
    try:
        keep_us = int(keep_us)
    except (ValueError, TypeError):
        raise ConfigurationError("invalid keep_us, must be an integer")
    if keep_us < -1:
        raise ConfigurationError("invalid keep_us, must be -1 (KEEP_ALL) or >= 0")
    return keep_us




def validate_name(name: str) -> str:
    if name is None or len(name) == 0:
        raise ConfigurationError("missing name")
    if '/' in name:
        raise ConfigurationError("invalid name, '\\' not allowed")
    return name



from click.testing import CliRunner
import warnings
import os
from datetime import datetime, timezone

from joule.models import DataStream, Element, StreamInfo
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

STREAM_INFO = os.path.join(os.path.dirname(__file__), 'stream.json')
warnings.simplefilter('always')


class TestStreamDelete(FakeJouleTestCase):
    def test_deletes_stream(self):
        server = FakeJoule()
        with open(STREAM_INFO, 'r') as f:
            server.response = f.read()
        server.stub_stream_info = True  # use the response text
        self.start_server(server)
        runner = CliRunner()
        # does not delete with out confirmation
        runner.invoke(main,
                      ['stream', 'delete',
                       '/folder/stream'],
                      input='\n')
        self.assertTrue(self.msgs.empty())

        # executes deletes with confirmation
        result = runner.invoke(main,
                               ['stream', 'delete',
                                '/folder/stream'],
                               input='y\n')
        self.assertEqual(result.exit_code, 0)
        deleted_stream = self.msgs.get()
        self.assertEqual(deleted_stream, '/folder/stream')
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        error_msg = "test error"
        error_code = 500
        server.response = error_msg
        server.stub_stream_destroy = True
        server.http_code = error_code
        # actually create a stream so the stubbed API call is the delete one
        src = DataStream(id=0, name="source", keep_us=100,
                         datatype=DataStream.DATATYPE.FLOAT32, updated_at=datetime.now(timezone.utc))
        src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]
        src_info = StreamInfo(0, 0, 0)
        server.add_stream('/folder/stream', src, src_info, None, [])

        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'delete', '/folder/stream'],
                               input='y\n')
        self.assertIn('%d' % error_code, result.output)
        self.assertIn(error_msg, result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



#!/usr/bin/env python3
import time
import re
import subprocess
import numpy as np
import asyncio

from joule import api
from joule.api.data_stream import DataStream, Element

def main():
    time.sleep(1)  # wait for jouled to boot
    asyncio.run(_run())


async def _run():
    node = api.get_node()
    procs = start_standalone_procs1()

    time.sleep(4)  # let procs run
    stop_standalone_procs(procs)

    time.sleep(8)  # close sockets
    procs = await start_standalone_procs2(node)
    time.sleep(4)  # let procs run
    stop_standalone_procs(procs)
    time.sleep(4)  # close sockets

    await check_streams(node)
    await check_modules(node)
    await check_data(node)
    await check_logs(node)
    await node.close()
    return


def start_standalone_procs1():
    # proc1 reads /counting/base and writes to /counting/plus3
    p1 = subprocess.Popen(build_standalone_args("proc1"))
    time.sleep(1)

    # proc1a reads /counting/base and writes to /counting/plus3
    #   fails because proc1 is already writing to path
    p1a = subprocess.run(build_standalone_args("proc1"),
                         stdout=subprocess.PIPE,
                         stderr=subprocess.PIPE,
                         universal_newlines=True)
    assert p1a.stderr.find("plus3") != -1
    # proc2 tries to read from /bad/path but fails
    p2 = subprocess.run(build_standalone_args("proc2"),
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        universal_newlines=True)
    assert p2.stderr.find("not being produced") != -1
    return [p1]


async def start_standalone_procs2(node: api.BaseNode):
    # proc3 tries to write to /counting/plus3 with wrong dtype
    p3 = subprocess.run(build_standalone_args("proc3"),
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        universal_newlines=True,
                        timeout=3)
    assert p3.stderr.find("layout") != -1

    #  proc1 reads /counting/base and writes to /counting/plus3
    p1 = subprocess.Popen(build_standalone_args("proc1"))

    stream = DataStream()
    stream.name = "int16"
    stream.datatype = "int16"
    for x in range(3):
        e = Element()
        e.name = 'item%d' % x
        e.index = x
        stream.elements.append(e)
    await node.data_stream_create(stream, "/exists")

    # proc5 tries to write to /exists/int16 with wrong element count
    p4 = subprocess.run(build_standalone_args("proc4"),
                        stdout=subprocess.PIPE,
                        stderr=subprocess.PIPE,
                        universal_newlines=True)
    assert p4.stderr.find("layout") != -1

    return [p1]


def build_standalone_args(proc_dir):
    scenario_dir = "/joule/tests/e2e/scenarios/standalone_modules/"
    base_dir = scenario_dir + "standalone_modules/" + proc_dir + "/"
    return ["coverage","run","--rcfile=/joule/.coveragerc",
            "/joule/tests/e2e/module_scripts/adder.py",
            "3", "--live",
            "--module_config", base_dir + "module.conf",
            "--stream_configs", base_dir + "stream_configs"]


def stop_standalone_procs(procs):
    for proc in procs:
        proc.kill()


async def check_modules(node: api.BaseNode):
    """
    Test: check module status
    Goal:
      counter: running some nonzero memory
      plus1: running some nonzero memory
    """
    modules = await node.module_list()
    assert (len(modules) == 2)


async def check_streams(node: api.BaseNode):
    """
    Test: check streams created by standalone modules
    Goal:
        /counting/plus3 should have customized display_type and keep settings
    """
    plus3 = await node.data_stream_get("/counting/plus3")
    assert (plus3.keep_us == (5 * 24 * 60 * 60 * 1e6))  # keep=5d
    assert (plus3.elements[0].display_type == "DISCRETE")


async def check_data(node: api.BaseNode):
    """
    Test: check data inserted into correctly
    Goal:
      /counting/base is int32_1, 1 interval with >300 samples
      /counting/plus1 is int32_1, data is base+1
      /counting/plus3 is int32_1, data is base+3

      all paths have decimations
    """
    base_path = "/counting/base"
    plus1_path = "/counting/plus1"
    plus3_path = "/counting/plus3"
    for path in [base_path, plus1_path]:
        # read 50 *live* samples from each stream
        pipe = await node.data_read(path)
        assert pipe.layout == "int32_1"
        num_intervals = 1
        len_data = 0
        while await pipe.not_empty():
            data = await pipe.read()
            len_data += len(data)
            if len_data > 50:
                break
            if pipe.end_of_interval:
                num_intervals += 1
            pipe.consume(len(data))

        # streams should have single continuous interval
        assert num_intervals == 1, "%s has %d intervals" % (path, num_intervals)
        assert len_data > 50
        await pipe.close()

    # verify plus3_path has 2 intervals
    intervals = await node.data_intervals(plus3_path)
    assert len(intervals) == 2, intervals

    # verify the filter module executed correctly
    # use the filter time bounds because to extract
    # the data to compare
    stream_info = await node.data_stream_info(plus1_path)
    p = await node.data_read(base_path,
                             start=stream_info.start,
                             end=stream_info.end,
                             max_rows=50)
    expected_data = await p.read(True)
    assert 0 < len(expected_data) <= 50
    expected_data[:, 1:] += 1
    await p.close()
    p = await node.data_read(plus1_path,
                             start=stream_info.start,
                             end=stream_info.end,
                             max_rows=50)
    actual_data = await p.read(True)
    verify_len = min(len(actual_data), len(expected_data))
    np.testing.assert_almost_equal(actual_data[:verify_len, :],
                                   expected_data[:verify_len, :])


async def check_logs(node: api.BaseNode):
    """
    Test: logs should contain info and stderr from modules
    Goal: 
      counter: says "starting" somewhere once
      plus1: says "starting" somewhere once
    """
    for module_name in ["counter", "plus1"]:
        logs = await node.module_logs(module_name)
        num_starts = len(re.findall("---starting module---", ' '.join(logs)))
        assert (num_starts == 1)


if __name__ == "__main__":
    main()
    print("OK")



from aiohttp import web
import json
from sqlalchemy.orm import Session
import datetime
from joule.models import EventStream, EventStore, event_stream
from joule.constants import ApiErrorMessages
from joule.controllers.helpers import read_json, validate_query_parameters
from joule.models import folder, Folder
from joule.errors import ConfigurationError
from joule import app_keys

async def info(request: web.Request):
    db: Session = request.app[app_keys.db]
    event_store: EventStore = request.app[app_keys.event_store]
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db, stream_type=EventStream)
    elif 'id' in request.query:
        my_stream = db.get(EventStream, request.query["id"])
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)
    if my_stream is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)
    stream_info = await event_store.info([my_stream])
    ## NOTE: this endpoint supports two separate API calls- event_stream_get and event_stream_info
    # The JSON response has data for both
    return web.json_response(my_stream.to_json(stream_info))


async def move(request: web.Request):
    db: Session = request.app[app_keys.db]
    body = await read_json(request)
    # find the stream
    if 'src_path' in body:
        my_stream = folder.find_stream_by_path(body['src_path'], db, stream_type=EventStream)
    elif 'src_id' in body:
        my_stream = db.get(EventStream, body["src_id"])
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)
    if my_stream is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)
    # find or create the destination folder
    if 'dest_path' in body:
        try:
            destination = folder.find(body['dest_path'], db, create=True)
        except ConfigurationError as e:
            raise web.HTTPBadRequest(reason="Destination error: %s" % str(e))
    elif 'dest_id' in body:
        destination = db.get(Folder, body["dest_id"])
    else:
        raise web.HTTPBadRequest(reason="specify a destination")
    # make sure name is unique in this destination
    existing_names = [s.name for s in destination.data_streams + destination.event_streams]
    if my_stream.name in existing_names:
        db.rollback()
        raise web.HTTPBadRequest(reason="stream with the same name exists in the destination folder")
    my_stream.folder.touch()
    destination.event_streams.append(my_stream)
    destination.touch()
    db.commit()
    return web.json_response({"stream": my_stream.to_json()})


async def create(request):
    db: Session = request.app[app_keys.db]
    event_store: EventStore = request.app[app_keys.event_store]
    body = await read_json(request)

    if 'stream' not in body:
        raise web.HTTPBadRequest(reason="provide a stream")

    # find or create the destination folder
    if 'dest_path' in body:
        try:
            destination = folder.find(body['dest_path'], db, create=True)
        except ConfigurationError as e:
            raise web.HTTPBadRequest(reason="Destination error: %s" % str(e))
    elif 'dest_id' in body:
        destination = db.get(Folder, body["dest_id"])
    else:
        raise web.HTTPBadRequest(reason="specify a destination")

    try:
        new_stream = event_stream.from_json(body['stream'])
        # clear out the id's
        new_stream.id = None
        # make sure name is unique in this destination
        existing_names = [s.name for s in destination.data_streams + destination.event_streams]
        if new_stream.name in existing_names:
            raise ConfigurationError("stream with the same name exists in the folder")
        destination.event_streams.append(new_stream)
        new_stream.touch()
        db.commit()
    except (TypeError, ValueError) as e:
        db.rollback()
        return web.Response(text="Invalid stream JSON: %r" % e, status=400)
    except ConfigurationError as e:
        db.rollback()
        return web.Response(text="Invalid stream specification: %s" % e, status=400)
    except KeyError as e:
        db.rollback()
        return web.Response(text="Invalid or missing stream attribute: %s" % e, status=400)
    await event_store.create(new_stream)
    return web.json_response(data=new_stream.to_json())


async def update(request: web.Request):
    db: Session = request.app[app_keys.db]
    body = await read_json(request)
    if 'id' not in body or body['id'] is None:
        raise web.HTTPBadRequest(reason="Invalid request: specify id")

    my_stream: EventStream = db.get(EventStream, body['id'])
    if my_stream is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)
    if 'stream' not in body:
        raise web.HTTPBadRequest(reason="Invalid request: specify stream as JSON")
    try:
        attrs = dict(body['stream'])
    except ValueError:
        raise web.HTTPBadRequest(reason="error: [stream] attribute must be JSON")
    try:
        my_stream.update_attributes(attrs)
        # make sure name is unique in this destination
        existing_names = [s.name for s in
                          my_stream.folder.data_streams + my_stream.folder.event_streams
                          if s.id != my_stream.id]
        if my_stream.name in existing_names:
            raise ConfigurationError("stream with the same name exists in the folder")
        db.commit()
    except (ValueError, ConfigurationError) as e:
        db.rollback()
        raise web.HTTPBadRequest(reason="Invalid stream specification: %s" % e)
    return web.json_response(data=my_stream.to_json())


async def delete(request):
    db: Session = request.app[app_keys.db]
    data_store: EventStore = request.app[app_keys.event_store]
    # find the requested stream
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db,
                                               stream_type=EventStream)
    elif 'id' in request.query:
        my_stream = db.get(EventStream, request.query["id"])
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)
    if my_stream is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)
    await data_store.destroy(my_stream)
    my_stream.folder.touch()
    db.delete(my_stream)
    db.commit()
    return web.Response(text="ok")


# ----- data actions ----

async def write_events(request):
    db: Session = request.app[app_keys.db]
    event_store: EventStore = request.app[app_keys.event_store]
    body = await read_json(request)

    # find the requested stream
    if 'path' in body:
        my_stream = folder.find_stream_by_path(body['path'], db,
                                               stream_type=EventStream)
    elif 'id' in body:
        my_stream = db.get(EventStream, body["id"])
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)
    if my_stream is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)
    if 'events' not in body:
        raise web.HTTPBadRequest(reason="specify events to add")
    # make sure the events belong to this stream, if event_stream_id does not match
    # this stream's ID, create a new event, otherwise copying events to a new stream
    # may lead to updating existing events (overwriting them) if the ID's collide
    events = body['events']
    for event in events:
        if 'event_stream_id' not in event:
            raise web.HTTPBadRequest(reason="all events must have an event_stream_id")
        if event['event_stream_id'] != my_stream.id:
            # this must have been copied from another stream, it is a new event, not an update
            event['id'] = None
            event['event_stream_id'] = my_stream.id
    events = await event_store.upsert(my_stream, events)
    return web.json_response(data={'count': len(events), 'events': events})


async def count_events(request):
    db: Session = request.app[app_keys.db]
    event_store: EventStore = request.app[app_keys.event_store]
    # find the requested stream
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db,
                                               stream_type=EventStream)
    elif 'id' in request.query:
        my_stream = db.get(EventStream, request.query["id"])
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)
    if my_stream is None:
        raise web.HTTPNotFound(reason="stream does not exist")

    # parse optional parameters
    params = {'start': None, 'end': None, 'include-ongoing-events': 0}
    validate_query_parameters(params, request.query)
    json_filter = _parse_json_filter(request.query)
    
    event_count = await event_store.count(my_stream, params['start'], params['end'],
                                          json_filter=json_filter,
                                          include_on_going_events=params['include-ongoing-events'])
    return web.json_response(data={'count': event_count})


async def read_events(request):
    db: Session = request.app[app_keys.db]
    event_store: EventStore = request.app[app_keys.event_store]
    # find the requested stream
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db,
                                               stream_type=EventStream)
    elif 'id' in request.query:
        my_stream = db.get(EventStream, request.query["id"])
    else:
        raise web.HTTPBadRequest(reason="specify an id or a path")
    if my_stream is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)
    if 'limit' not in request.query:
        raise web.HTTPBadRequest(reason="limit parameter is required")
    try:
        limit = int(request.query['limit'])
        if limit < 0:
            raise ValueError
    except ValueError:
        raise web.HTTPBadRequest(reason="limit parameter must be an integer > 0")
    params = {'start': None, 'end': None, 'include-ongoing-events': 0}
    validate_query_parameters(params, request.query)
    json_filter = _parse_json_filter(request.query)
    
    # handle limit parameter, default is HARD, do not return unless count < limit
    if 'return-subset' not in request.query:
        event_count = await event_store.count(my_stream, params['start'], params['end'],
                                              json_filter=json_filter,
                                              include_on_going_events=params['include-ongoing-events'])
        if event_count > limit:
            # too many events, send a histogram instead
            event_hist = await event_store.histogram(my_stream, params['start'], params['end'],
                                                     json_filter=json_filter,
                                                     nbuckets=100)
            return web.json_response(data={'count': event_count, 'events': event_hist, 'type': 'histogram'})
    # if return-subset, limit is SOFT, return just that many events
    events = await event_store.extract(my_stream, params['start'], params['end'],
                                       limit=limit, json_filter=json_filter,
                                       include_on_going_events=params['include-ongoing-events'])
    return web.json_response(data={'count': len(events), 'events': events, 'type': 'events'})


async def remove_events(request):
    db: Session = request.app[app_keys.db]
    event_store: EventStore = request.app[app_keys.event_store]
    # find the requested stream
    if 'path' in request.query:
        my_stream = folder.find_stream_by_path(request.query['path'], db,
                                               stream_type=EventStream)
    elif 'id' in request.query:
        my_stream = db.get(EventStream, request.query["id"])
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)
    if my_stream is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.stream_does_not_exist)

    # parse optional parameters
    params = {'start': None, 'end': None}
    validate_query_parameters(params, request.query)
    json_filter = _parse_json_filter(request.query)
    
    await event_store.remove(my_stream, params['start'], params['end'], json_filter=json_filter)
    return web.Response(text="ok")

def _parse_json_filter(query):
    # handle json filter parameter
    json_filter = None
    if 'filter' in query and query['filter'] is not None and len(query['filter']) > 0:
        try:
            json_filter = json.loads(query['filter'])
            # TODO verify syntax
        except (json.decoder.JSONDecodeError, ValueError):
            raise web.HTTPBadRequest(reason=ApiErrorMessages.invalid_filter_parameter)
    return json_filter



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
import aiohttp

import joule.controllers
from joule.models.supervisor import Supervisor
from tests.controllers.helpers import MockWorker
from joule import app_keys
from joule.constants import EndPoints

class TestModuleControllerErrors(AioHTTPTestCase):

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        wreader = MockWorker("reader", {}, {'output': '/reader/path'})
        wfilter = MockWorker("filter", {'input': '/reader/path'}, {'output': '/output/path'})
        app[app_keys.supervisor] = Supervisor([wreader, wfilter], [], None)  # type: ignore
        return app


    async def test_module_info(self):
        # must specify a name
        resp: aiohttp.ClientResponse = await self.client.request("GET", EndPoints.module)
        self.assertEqual(resp.status, 400)
        # return "not found" on bad name
        resp: aiohttp.ClientResponse = await self.client.request("GET", EndPoints.module,
                                                                 params={'name': 'unknown'})
        self.assertEqual(resp.status, 404)


    async def test_module_logs(self):
        # must specify a name
        resp: aiohttp.ClientResponse = await self.client.request("GET", EndPoints.module_logs)
        self.assertEqual(resp.status, 400)
        # return "not found" on bad name
        resp: aiohttp.ClientResponse = await self.client.request("GET", EndPoints.module_logs,
                                                                 params={'name': 'unknown'})
        self.assertEqual(resp.status, 404)



from click.testing import CliRunner
import os
from aiohttp.test_utils import unused_port
import warnings
import json
import copy
import asyncio
import numpy as np

from tests import helpers

from joule.models import DataStream, Element, StreamInfo, pipes
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

warnings.simplefilter('always')


class TestDataIntervals(FakeJouleTestCase):

    def test_shows_all_data_intervals(self):
        server = FakeJoule()
        intervals = create_source_data(server)
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'intervals',
                                      '--start', '1 hour ago', '--end','now',
                                      '/test/source'])

        self.assertEqual(result.exit_code, 0)
        # make sure the intervals are displayed (one per line)
        output = result.output.split('\n')
        num_intervals = 0
        for line in output[1:]:
            if '[' in line:
                self.assertIn('Thu, 24 Jan', line)
                num_intervals+=1
        self.assertEqual(num_intervals, len(intervals))
        self.stop_server()

    def test_handles_no_intervals(self):
        server = FakeJoule()
        create_source_data(server, no_intervals=True)
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'intervals',
                                      '--start', '1 hour ago', '--end', 'now',
                                      '/test/source'])
        self.assertEqual(result.exit_code, 0)
        # make sure the intervals are displayed (one per line)
        output = result.output.split('\n')
        for line in output[1:]:
            self.assertNotIn('[', line)

        self.stop_server()

    def test_when_stream_does_not_exist(self):
        server = FakeJoule()
        server.response = "stream does not exist"
        server.http_code = 404
        server.stub_data_intervals = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'intervals', '/bad/path'])
        self.assertIn("Error", result.output)
        self.stop_server()

    def test_when_server_returns_invalid_data(self):
        server = FakeJoule()
        server.response = "notjson"
        server.stub_data_intervals = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'intervals', 'folder/stream'])
        self.assertIn('Error', result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        server.response = "test error"
        server.http_code = 500
        server.stub_data_intervals = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['data', 'intervals', 'folder/stream'])
        self.assertIn('500', result.output)
        self.assertIn("test error", result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()


def create_source_data(server, no_intervals=False):
    # create the source stream
    src = DataStream(id=0, name="source", keep_us=100, datatype=DataStream.DATATYPE.FLOAT32)
    src.elements = [Element(name="e%d" % x, index=x, display_type=Element.DISPLAYTYPE.CONTINUOUS) for x in range(3)]

    # source has 100 rows of data in four intervals between [0, 100]
    src_data = helpers.create_data(src.layout, length=100, start=1548353881*1e6, step=1e6)

    ts = src_data['timestamp']

    if no_intervals:
        intervals = []
        src_info = StreamInfo(None, None, 0)

    else:
        intervals = [[ts[0], ts[24]],
                     [ts[25], ts[49]],
                     [ts[50], ts[74]],
                     [ts[75], ts[99]]]
        src_info = StreamInfo(intervals[0][0], intervals[-1][1], len(src_data))

    server.add_stream('/test/source', src, src_info, src_data, intervals)
    return intervals



from aiohttp import web
from sqlalchemy.exc import CircularDependencyError
from sqlalchemy.orm import Session
from typing import List

from joule.constants import ApiErrorMessages
from joule.models import folder, Folder, EventStore, DataStore, DataStream, EventStream
from joule.errors import ConfigurationError
from joule import app_keys

async def index(request: web.Request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    event_store: DataStore = request.app[app_keys.event_store]

    root = folder.root(db)
    if 'data-info' in request.query:
        data_streams = db.query(DataStream).all()
        data_streams_info = await data_store.info(data_streams)
        event_streams = db.query(EventStream).all()
        event_streams_info = await event_store.info(event_streams)
    else:
        data_streams_info = {}
        event_streams_info = {}
    resp = root.to_json(data_streams_info, event_streams_info)
    # get a list of the ids of all DataStreams that are either a source or a destination
    query_result = db.query(DataStream). \
        filter((DataStream.is_source == True) | (DataStream.is_destination == True)) \
        .with_entities(DataStream.id).all()
    resp['active_data_streams'] = [r[0] for r in query_result]
    return web.json_response(resp)


async def info(request: web.Request):
    db: Session = request.app[app_keys.db]
    if 'path' in request.query:
        my_folder = folder.find(request.query['path'], db)
    elif 'id' in request.query:
        my_folder = db.get(Folder, request.query["id"])
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)
    if my_folder is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.folder_does_not_exist)
    return web.json_response(my_folder.to_json())


async def move(request):
    db: Session = request.app[app_keys.db]
    if request.content_type != 'application/json':
        raise web.HTTPBadRequest(reason='content-type must be application/json')
    body = await request.json()
    # find the folder
    if 'src_path' in body:
        my_folder = folder.find(body['src_path'], db)
    elif 'src_id' in body:
        my_folder = db.get(folder.Folder, body["src_id"])
    else:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.specify_id_or_path)
    if my_folder is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.folder_does_not_exist)
    # make sure the folder is not locked
    if my_folder.locked:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.folder_is_locked)
    # find or create the destination folder
    if 'dest_path' in body:
        try:
            destination = folder.find(body['dest_path'], db, create=True)
        except ConfigurationError as e:
            raise web.HTTPBadRequest(reason="Destination error: %s" % str(e))
    elif 'dest_id' in body:
        destination = db.get(Folder, body["dest_id"])
    else:
        raise web.HTTPBadRequest(reason="specify a destination")
    
    # make sure name is unique in the folder
    peer_names = [f.name for f in destination.children]
    if my_folder.name in peer_names:
        db.rollback()
        raise web.HTTPBadRequest(reason="a folder with this name is in the destination folder")
    try:
        # make sure this does not create a circular graph
        parent = destination.parent
        while parent is not None:
            if parent == my_folder:
                raise ValueError()
            parent = parent.parent
        my_folder.touch()
        destination.touch()
        destination.children.append(my_folder)
        db.commit()
    except (CircularDependencyError, ValueError):
        db.rollback()
        raise web.HTTPBadRequest(reason="cannot place a parent in a child folder")
    return web.json_response({"stream": my_folder.to_json()})


async def update(request):
    db: Session = request.app[app_keys.db]
    if request.content_type != 'application/json':
        raise web.HTTPBadRequest(reason='content-type must be application/json')
    body = await request.json()
    if 'id' not in body:
        raise web.HTTPBadRequest(reason="Invalid request: specify id")

    my_folder: Folder = db.get(Folder, body['id'])
    if my_folder is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.folder_does_not_exist)
    if my_folder.locked:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.folder_is_locked)
    if 'folder' not in body:
        raise web.HTTPBadRequest(reason="Invalid request: specify folder as JSON")
    try:
        attrs = dict(body['folder'])
    except ValueError:
        raise web.HTTPBadRequest(reason="error: [folder] attribute must be JSON")
    try:
        my_folder.update_attributes(attrs)
        # make sure name is unique in the folder
        peer_names = [f.name for f in my_folder.parent.children if f.id != my_folder.id]
        if my_folder.name in peer_names:
            db.rollback()
            return web.Response(text="a folder with this name is in the destination folder",
                                status=400)
        my_folder.touch()
        db.commit()
    except (ValueError, ConfigurationError) as e:
        db.rollback()
        raise web.HTTPBadRequest(reason="error: %s" % e)
    return web.json_response(data=my_folder.to_json())


async def delete(request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    event_store: EventStore = request.app[app_keys.event_store]
    # find the requested folder
    if 'path' in request.query:
        my_folder: Folder = folder.find(request.query['path'], db)
    elif 'id' in request.query:
        my_folder = db.get(Folder, request.query["id"])
    else:
        raise web.HTTPBadRequest(reason="specify an id or a path")
    if my_folder is None:
        raise web.HTTPNotFound(reason=ApiErrorMessages.folder_does_not_exist)
    if my_folder.locked:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.folder_is_locked)
    if 'recursive' in request.query and request.query["recursive"] == '1':
        await _recursive_delete(my_folder.children, db, data_store, event_store)
    elif len(my_folder.children) > 0:
        raise web.HTTPBadRequest(reason="specify [recursive] or remove child folders first")

    for stream in my_folder.data_streams:
        if stream.locked:  
            # shouldn't ever get here because of the check above, but just in case
            raise web.HTTPBadRequest(reason="cannot delete folder with locked streams") # pragma: no cover
        await data_store.destroy(stream)
        db.delete(stream)
    for stream in my_folder.event_streams:
        await event_store.destroy(stream)
        db.delete(stream)

    db.delete(my_folder)
    # update the parent timestamps
    my_folder.parent.touch()
    db.commit()
    return web.Response(text="ok")


async def _recursive_delete(folders: List[Folder], db: Session, data_store: DataStore,
                            event_store: EventStore):
    if len(folders) == 0:
        return
    for f in folders:
        for stream in f.data_streams:
            if stream.locked:  
                # shouldn't ever get here because a locked folder shouldn't call us
                raise web.HTTPBadRequest(reason="cannot delete folder with locked streams") # pragma: no cover
            await data_store.destroy(stream)
            db.delete(stream)
        for stream in f.event_streams:
            await event_store.destroy(stream)
            db.delete(stream)
        await _recursive_delete(f.children, db, data_store, event_store)
        db.delete(f)



from click.testing import CliRunner
import os
import warnings
import json
import copy
import asyncio
import re

from joule.models import DataStream
from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

STREAM_INFO = os.path.join(os.path.dirname(__file__), 'stream.json')
warnings.simplefilter('always')


class TestStreamInfo(FakeJouleTestCase):

    def test_shows_extended_stream_info(self):
        server = FakeJoule()
        with open(STREAM_INFO, 'r') as f:
            server.response = f.read()
        server.stub_stream_info = True  # use the response text
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'info', '-e', '/folder_1/random'])
        self.assertEqual(result.exit_code, 0)
        # make sure the items are populated correctly
        output = result.output.split('\n')
        row_line = [x for x in output if 'Rows' in x][0]
        self.assertIn("73820", row_line)
        start_line = [x for x in output if 'Start' in x][0]

        # note: hour depends on client's timezone

        self.assertIsNotNone(re.search(r'2018-07-11 \d\d:50:44',
                                       start_line))
        end_line = [x for x in output if 'End' in x][0]
        self.assertIsNotNone(re.search(r'2018-07-25 \d\d:52:56',
                                       end_line))
        for element_name in ['x', 'y', 'z']:
            for line in output:
                if element_name in line:
                    break
            else:
                self.fail("element name %s not in output" % element_name)
        self.stop_server()

    def test_shows_basic_stream_info(self):
        server = FakeJoule()
        with open(STREAM_INFO, 'r') as f:
            server.response = f.read()
        server.stub_stream_info = True  # use the response text
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'info', '/folder_1/random'])
        self.assertEqual(result.exit_code, 0)
        # make sure the items are populated correctly
        output = result.output.split('\n')
        row_line = [x for x in output if 'Rows' in x][0]
        self.assertIn("73820", row_line)
        start_line = [x for x in output if 'Start' in x][0]

        # note: hour depends on client's timezone

        self.assertIsNotNone(re.search(r'2018-07-11 \d\d:50:44',
                                       start_line))
        end_line = [x for x in output if 'End' in x][0]
        self.assertIsNotNone(re.search(r'2018-07-25 \d\d:52:56',
                                       end_line))
        # no table of rows
        for line in output:
            if 'Min,Max' in line:
                self.fail("it looks like the elements are still included in the display")
        self.stop_server()

    ## test alternate configurations by adjusting the JSON response

    def test_shows_when_stream_is_active(self):
        server = FakeJoule()
        with open(STREAM_INFO, 'r') as f:
            stream_data = json.loads(f.read())
        stream_data['active'] = True
        server.stub_stream_info = True  # use the response text
        server.response = json.dumps(stream_data)
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'info', '/folder_1/random'])
        self.assertEqual(result.exit_code, 0)
        # make sure the items are populated correctly
        self.assertIn("active", result.output)
        self.stop_server()

    def test_shows_when_stream_is_idle(self):
        server = FakeJoule()
        with open(STREAM_INFO, 'r') as f:
            stream_data = json.loads(f.read())
        stream_data['active'] = False
        stream_data['locked'] = False
        server.stub_stream_info = True  # use the response text
        server.response = json.dumps(stream_data)
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'info', '/folder_1/random'])
        self.assertEqual(result.exit_code, 0)
        # make sure the items are populated correctly
        self.assertIn("idle", result.output)
        self.stop_server()


    def test_handles_different_stream_configurations(self):
        server = FakeJoule()
        with open(STREAM_INFO, 'r') as f:
            orig_stream_data = json.loads(f.read())
        stream1 = copy.deepcopy(orig_stream_data)
        stream1['decimate'] = False
        # stream1 is configured but not active
        stream1['is_source'] = False
        stream1['is_destination'] = False
        stream1['is_configured'] = True
        stream1['keep_us'] = 8*60*60*1e6  # 8 hours
        stream2 = copy.deepcopy(orig_stream_data)
        stream2['keep_us'] = DataStream.KEEP_NONE
        stream2['description'] = 'description'
        stream2['data_info']['start'] = None
        stream2['data_info']['end'] = None
        # stream2 is idle
        stream2['is_source'] = False
        stream2['is_destination'] = False
        stream2['is_configured'] = False
        for stream in [stream1, stream2]:


            server.stub_stream_info = True  # use the response text
            server.response = json.dumps(stream)
            self.start_server(server)
            runner = CliRunner()
            result = runner.invoke(main, ['stream', 'info', '/folder_1/random'])
            # just make sure different configurations do not cause errors in the output
            self.assertEqual(result.exit_code, 0)
            self.stop_server()

    def test_when_stream_does_not_exist(self):
        server = FakeJoule()
        server.response = "stream does not exist"
        server.http_code = 404
        server.stub_stream_info = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, [ 'stream', 'info', '/bad/path'])
        self.assertIn("Error", result.output)
        self.stop_server()

    def test_when_server_returns_invalid_data(self):
        server = FakeJoule()
        server.response = "notjson"
        server.stub_stream_info = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'info', 'folder/stream'])
        self.assertIn('Error', result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        server.response = "test error"
        server.http_code = 500
        server.stub_stream_info = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['stream', 'info', 'folder/stream'])
        self.assertIn('500', result.output)
        self.assertIn("test error", result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



from aiohttp import web
from joule.constants import EndPoints

from joule.controllers import (
    root_controller,
    stream_controller,
    event_controller,
    folder_controller,
    data_controller,
    module_controller,
    app_controller,
    proxy_controller,
    master_controller,
    follower_controller,
    annotation_controller)

routes = [
    web.get(EndPoints.root, root_controller.index),
    web.get(EndPoints.db_connection, root_controller.db_connection),
    web.get(EndPoints.db_info, root_controller.dbinfo),
    web.get(EndPoints.version, root_controller.version),
    web.get(EndPoints.version_json, root_controller.version_json),
    # --- event stream routes ---
    web.get(EndPoints.event, event_controller.info),
    web.put(EndPoints.event_move, event_controller.move),
    web.put(EndPoints.event, event_controller.update),
    web.post(EndPoints.event, event_controller.create),
    web.delete(EndPoints.event, event_controller.delete),
    # --- event stream data routes ---
    web.get(EndPoints.event_data, event_controller.read_events),
    web.get(EndPoints.event_data_count, event_controller.count_events),
    web.post(EndPoints.event_data, event_controller.write_events),
    web.delete(EndPoints.event_data, event_controller.remove_events),
    # --- stream routes ---
    web.get(EndPoints.stream, stream_controller.info),
    web.put(EndPoints.stream_move, stream_controller.move),
    web.put(EndPoints.stream, stream_controller.update),
    web.post(EndPoints.stream, stream_controller.create),
    web.delete(EndPoints.stream, stream_controller.delete),
    # --- folder routes ---
    web.get(EndPoints.folders, folder_controller.index),
    web.get(EndPoints.folder, folder_controller.info),
    web.put(EndPoints.folder_move, folder_controller.move),
    web.put(EndPoints.folder, folder_controller.update),
    web.delete(EndPoints.folder, folder_controller.delete),
    # --- data routes ---
    web.get(EndPoints.data, data_controller.read),
    web.get(EndPoints.data_json, data_controller.read_json),
    web.get(EndPoints.data_intervals, data_controller.intervals),
    web.post(EndPoints.data_decimate, data_controller.decimate),
    web.delete(EndPoints.data_decimate, data_controller.drop_decimations),
    web.post(EndPoints.data_consolidate, data_controller.consolidate),
    web.post(EndPoints.data, data_controller.write),
    web.delete(EndPoints.data, data_controller.remove),
    # --- module routes ---
    web.get(EndPoints.modules, module_controller.index),
    web.get(EndPoints.module, module_controller.info),
    web.get(EndPoints.module_logs, module_controller.logs),
    # --- app routes ---
    web.get(EndPoints.app_auth, app_controller.auth),
    web.get(EndPoints.app_json, app_controller.index),

    # --- proxy routes ---
    web.get(EndPoints.proxies, proxy_controller.index),
    web.get(EndPoints.proxy, proxy_controller.info),
    # -- master routes --
    web.get(EndPoints.masters, master_controller.index),
    web.post(EndPoints.master, master_controller.add),
    web.delete(EndPoints.master, master_controller.delete),
    # -- follower routes --
    web.get(EndPoints.followers, follower_controller.index),
    web.post(EndPoints.follower, follower_controller.add),
    web.delete(EndPoints.follower, follower_controller.delete),
    # -- annotation routes --
    web.get(EndPoints.annotations_info, annotation_controller.info),
    web.get(EndPoints.annotations, annotation_controller.index),
    web.put(EndPoints.annotation, annotation_controller.update),
    web.post(EndPoints.annotation, annotation_controller.create),
    web.delete(EndPoints.annotation, annotation_controller.delete),
    web.delete(EndPoints.stream_annotations, annotation_controller.delete_all)
]

insecure_routes = [
    ['POST', EndPoints.follower]
]



from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web
import aiohttp
from sqlalchemy.orm import Session
import asyncio
from joule.models import DataStream
import joule.controllers
from tests.controllers.helpers import create_db, MockStore, MockSupervisor
from tests import helpers
from joule import app_keys
from joule.constants import EndPoints

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestDataController(AioHTTPTestCase):

    async def tearDownAsync(self):
        self.app[app_keys.db].close()
        self.app[psql_key].stop()
        await self.client.close()

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)
        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0
        app[app_keys.db], app[psql_key] = create_db(["/folder1/stream1:float32[x, y, z]",
                                            "/folder2/deeper/stream2:int16[val1, val2]"])
        app[app_keys.data_store] = MockStore()
        self.supervisor = MockSupervisor()
        app[app_keys.supervisor] = self.supervisor
        return app


    async def test_read_requires_path_or_id(self):
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data, params={})
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('path' in await resp.text())


    async def test_read_errors_on_invalid_stream(self):
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data, params={'path': '/no/stream'})
        self.assertEqual(resp.status, 404)
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data, params={'id': '404'})
        self.assertEqual(resp.status, 404)


    async def test_read_errors_on_no_data(self):
        store: MockStore = self.app[app_keys.data_store]
        store.configure_extract(0, no_data=True)
        params = {'path': '/folder1/stream1', 'decimation-level': 64}
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('no data' in await resp.text())


    async def test_read_errors_on_decimation_failure(self):
        store: MockStore = self.app[app_keys.data_store]
        store.configure_extract(10, decimation_error=True)
        params = {'path': '/folder1/stream1', 'decimation-level': 64}
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('decimated' in await resp.text())


    async def test_read_errors_on_datastore_error(self):
        store: MockStore = self.app[app_keys.data_store]
        store.configure_extract(10, data_error=True)
        params = {'path': '/folder1/stream1', 'decimation-level': 64}
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('error' in await resp.text())


    async def test_read_params_must_be_valid(self):
        params = dict(path='/folder1/stream1')
        # start must be an int
        params['start'] = 'bad'
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('start' in await resp.text())
        # end must be an int
        params['start'] = 100
        params['end'] = '200.5'
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('end' in await resp.text())
        # start must be before end
        params['end'] = 50
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('end' in await resp.text())
        self.assertTrue('start' in await resp.text())
        params['end'] = 200
        # max rows must be an int > 0
        params['max-rows'] = -4
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('max-rows' in await resp.text())
        params['max-rows'] = 100
        # decimation-level must be >= 0
        params['decimation-level'] = -50
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('decimation-level' in await resp.text())
        params['decimation-level'] = 20


    async def test_subscribe_requires_path_or_id(self):
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data, params={'subscribe': '1'})
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('path' in await resp.text())


    async def test_subscribe_errors_on_invalid_stream(self):
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data, params={'path': '/no/stream', 'subscribe': '1'})
        self.assertEqual(resp.status, 404)
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data, params={'id': '404', 'subscribe': '1'})
        self.assertEqual(resp.status, 404)


    async def test_subscribe_errors_on_unproduced_stream(self):
        self.supervisor.raise_error = True
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data, params={'path': '/folder1/stream1', 'subscribe': '1'})
        self.assertEqual(resp.status, 400)


    async def test_subscribe_does_not_support_json(self):
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data_json, params={'path': '/folder1/stream1', 'subscribe': '1'})
        self.assertEqual(resp.status, 400)


    async def test_write_requires_path_or_id(self):
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={})
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('path' in await resp.text())


    async def test_write_errors_on_invalid_stream(self):
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={'path': '/no/stream'})
        self.assertEqual(resp.status, 404)
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={'id': '404'})
        self.assertEqual(resp.status, 404)

    async def test_write_errors_on_merge_gap(self):
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={'path': '/folder1/stream1', 'merge-gap':'not valid'})
        self.assertEqual(resp.status, 400)
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={'path': '/folder1/stream1', 'merge-gap':-5})
        self.assertEqual(resp.status, 400)
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={'path': '/folder1/stream1', 'merge-gap':8.6})
        self.assertEqual(resp.status, 400)

    async def test_remove_requires_path_or_id(self):
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.data, params={})
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('path' in await resp.text())


    async def test_remove_errors_on_invalid_stream(self):
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.data, params={'path': '/no/stream'})
        self.assertEqual(resp.status, 404)
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.data, params={'id': '404'})
        self.assertEqual(resp.status, 404)


    async def test_remove_bounds_must_be_valid(self):
        params = dict(path='/folder1/stream1')
        # start must be an int
        params['start'] = 'bad'
        resp: aiohttp.ClientResponse = await self.client.delete(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('start' in await resp.text())
        # end must be an int
        params['start'] = 100
        params['end'] = '200.5'
        resp: aiohttp.ClientResponse = await self.client.delete(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('end' in await resp.text())
        # start must be before end
        params['end'] = 50
        resp: aiohttp.ClientResponse = await self.client.delete(EndPoints.data, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('end' in await resp.text())
        self.assertTrue('start' in await resp.text())
        params['end'] = 200


    async def test_intervals_requires_path_or_id(self):
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data_intervals, params={})
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('path' in await resp.text())


    async def test_intervals_errors_on_bad_stream(self):
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data_intervals, params={'path': '/no/stream'})
        self.assertEqual(resp.status, 404)
        resp: aiohttp.ClientResponse = await \
            self.client.get(EndPoints.data_intervals, params={'id': '404'})
        self.assertEqual(resp.status, 404)


    async def test_interval_bounds_must_be_valid(self):
        params = dict(path='/folder1/stream1')
        # start must be an int
        params['start'] = 'bad'
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data_intervals, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('start' in await resp.text())
        # end must be an int
        params['start'] = 100
        params['end'] = '200.5'
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data_intervals, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('end' in await resp.text())
        # start must be before end
        params['end'] = 50
        resp: aiohttp.ClientResponse = await self.client.get(EndPoints.data_intervals, params=params)
        self.assertNotEqual(resp.status, 200)
        self.assertTrue('end' in await resp.text())
        self.assertTrue('start' in await resp.text())
        params['end'] = 200


    async def test_invalid_writes_propagates_data_error(self):
        db: Session = self.app[app_keys.db]
        store: MockStore = self.app[app_keys.data_store]
        store.raise_data_error = True
        stream: DataStream = db.query(DataStream).filter_by(name="stream1").one()
        data = helpers.create_data(stream.layout)
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data, params={"path": "/folder1/stream1"},
                             data=data.tobytes())
        self.assertEqual(resp.status, 400)
        self.assertIn('test error', await resp.text())

    async def test_cannot_use_decimation_tools_without_backend_support(self):
        store: MockStore = self.app[app_keys.data_store]
        store.supports_decimation_management = False
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data_decimate, params={"path": "/folder1/stream1"})
        self.assertEqual(resp.status, 400)
        self.assertIn('support', await resp.text())
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data_consolidate, params={"path": "/folder1/stream1"})
        self.assertEqual(resp.status, 400)
        self.assertIn('support', await resp.text())
        resp: aiohttp.ClientResponse = await \
            self.client.delete(EndPoints.data_decimate, params={"path": "/folder1/stream1"})
        self.assertEqual(resp.status, 400)
        self.assertIn('support', await resp.text())

    async def test_data_stream_consolidation(self):
        store: MockStore = self.app[app_keys.data_store]
        store.supports_consolidation = True
        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data_consolidate, params={"path": "/folder1/stream1"})
        self.assertEqual(resp.status, 400)
        self.assertIn('max_gap', await resp.text())

        resp: aiohttp.ClientResponse = await \
            self.client.post(EndPoints.data_consolidate, params={"path": "/folder1/stream1", "max_gap": -2})
        self.assertEqual(resp.status, 400)
        self.assertIn('max_gap', await resp.text())


from click.testing import CliRunner
import os
import warnings
import re

from ..fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

EVENT_INFO = os.path.join(os.path.dirname(__file__), 'event_info.json')
EMPTY_EVENT_INFO = os.path.join(os.path.dirname(__file__), 'empty_event_info.json')

warnings.simplefilter('always')


class TestEventInfo(FakeJouleTestCase):

    def test_show_info(self):
        server = FakeJoule()
        with open(EVENT_INFO, 'r') as f:
            server.response = f.read()
        server.stub_event_info = True  # use the response text
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['event', 'info', '/folder_1/random'])
        self.assertCliSuccess(result)
        self.assertEqual(result.exit_code, 0)
        # make sure the items are populated correctly
        output = result.output.split('\n')
        row_line = [x for x in output if 'Events' in x][0]
        self.assertIn("282587", row_line)
        start_line = [x for x in output if 'Start' in x][0]

        # note: day and hour depends on client's timezone (check year and MM:SS)
        self.assertIsNotNone(re.search(r'2013-03-', start_line))
        end_line = [x for x in output if 'End' in x][0]
        self.assertIsNotNone(re.search(r'28:49', end_line))

        # make sure the event fields are present
        for field in ["type", "phase"]:
            line = [x for x in output if field in x][0]
            self.assertIn("string", line)
        for field in ['duration', 'isolated', 'position',
                      'real_power', 'power_factor', 'apparent_power',
                      'reactive_power', 'relative_variance']:
            line = [x for x in output if field in x][0]
            self.assertIn("numeric", line)
        self.stop_server()

    def test_shows_empty_info(self):
        server = FakeJoule()
        with open(EMPTY_EVENT_INFO, 'r') as f:
            server.response = f.read()
        server.stub_event_info = True  # use the response text
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['event', 'info', '/folder_1/random'])
        self.assertCliSuccess(result)
        self.assertEqual(result.exit_code, 0)
        # make sure the items are populated correctly
        output = result.output.split('\n')
        row_line = [x for x in output if 'Events' in x][0]
        self.assertIn("0", row_line)
        start_line = [x for x in output if 'Start' in x][0]

        # no time bounds
        self.assertIn(u"\u2014", start_line)
        end_line = [x for x in output if 'End' in x][0]
        self.assertIn(u"\u2014", end_line)

        # no event fields printed
        self.assertIn("no fields specified", result.output)
        self.stop_server()

    def test_errors_on_invalid_path(self):
        server = FakeJoule()
        server.response = "stream does not exist"
        server.http_code = 404
        server.stub_stream_info = True
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['event', 'info', '/bad/path'])
        self.assertIn("Error", result.output)
        self.stop_server()




from tests.api import mock_session
import unittest
from joule.api.node import TcpNode
from joule import errors

from joule.api.data_stream import DataStream
from joule.api.folder import Folder
from .helpers import build_stream
from joule.constants import EndPoints


class TestStreamApi(unittest.IsolatedAsyncioTestCase):

    async def asyncSetUp(self):
        # no URL or event loop
        self.node = TcpNode('mock_node', 'http://url', 'api_key')
        self.session = mock_session.MockSession()
        self.node.session = self.session

    async def test_stream_move(self):
        # can move by ID
        await self.node.data_stream_move(1, 2)
        self.assertEqual(self.session.method, 'PUT')
        self.assertEqual(self.session.path, EndPoints.stream_move)
        self.assertEqual(self.session.request_data,
                         {'src_id': 1, 'dest_id': 2})
        # can move by path
        await self.node.data_stream_move('/a/path', '/b/path')
        self.assertEqual(self.session.method, 'PUT')
        self.assertEqual(self.session.path, EndPoints.stream_move)
        self.assertEqual(self.session.request_data,
                         {'src_path': '/a/path', 'dest_path': '/b/path'})

        # can move by Folder and DataStream
        src = DataStream()
        src.id = 1
        dest = Folder()
        dest.id = 2
        await self.node.data_stream_move(src, dest)
        self.assertEqual(self.session.path, EndPoints.stream_move)
        self.assertEqual(self.session.request_data,
                         {'src_id': 1, 'dest_id': 2})

    async def test_stream_move_errors(self):
        self.session.method = None
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_move(DataStream(), [])
        self.assertIsNone(self.session.method)

        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_move([], Folder())
        self.assertIsNone(self.session.method)

    async def test_stream_delete(self):
        # can delete by ID
        await self.node.data_stream_delete(1)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.stream)
        self.assertEqual(self.session.request_data, {'id': 1})
        # can delete by path
        await self.node.data_stream_delete('/a/path')
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.stream)
        self.assertEqual(self.session.request_data, {'path': '/a/path'})

        # can delete by DataStream
        src = DataStream()
        src.id = 1
        await self.node.data_stream_delete(src)
        self.assertEqual(self.session.method, 'DELETE')
        self.assertEqual(self.session.path, EndPoints.stream)
        self.assertEqual(self.session.request_data, {'id': 1})

    async def test_stream_delete_errors(self):
        self.session.method = None
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_delete(Folder())
        self.assertIsNone(self.session.method)

    async def test_stream_get(self):
        target = build_stream('test', 'int16[a,b]')
        target.id = 1
        self.session.response_data = target.to_json()
        stream = await self.node.data_stream_get(1)
        self.assertEqual(stream.to_json(), target.to_json())
        self.assertEqual(self.session.method, 'GET')
        self.assertEqual(self.session.path, EndPoints.stream)
        self.assertEqual(self.session.request_data, {'id': 1, 'no-info':''})
        # can get by path
        await self.node.data_stream_get('/a/path')
        self.assertEqual(self.session.method, 'GET')
        self.assertEqual(self.session.path, EndPoints.stream)
        self.assertEqual(self.session.request_data, {'path': '/a/path', 'no-info':''})

        # can get by DataStream
        src = DataStream()
        src.id = 1
        await self.node.data_stream_get(src)
        self.assertEqual(self.session.path, EndPoints.stream)
        self.assertEqual(self.session.request_data, {'id': 1, 'no-info':''})

    async def test_stream_get_errors(self):
        self.session.method = None
        with self.assertRaises(errors.ApiError):
            await self.node.data_stream_get(Folder())
        self.assertIsNone(self.session.method)

    async def test_local_stream_has_no_id(self):
        stream = build_stream('test', 'float64[z,y]')
        with self.assertRaises(errors.ApiError):
            stream.id



from sqlalchemy.orm import Session
from aiohttp import web
import numpy as np
import logging
from joule import app_keys

from joule.models import (DataStore,
                          InsufficientDecimationError, DataError,
                          pipes)
from joule.models.supervisor import Supervisor
from joule.errors import SubscriptionError
from joule.controllers.helpers import validate_query_parameters, get_stream_from_request_params
log = logging.getLogger('joule')


async def read_json(request: web.Request):
    return await read(request, json=True)


async def read(request: web.Request, json=False):
    if 'subscribe' in request.query and request.query['subscribe'] == '1':
        return await _subscribe(request, json)
    else:
        return await _read(request, json)


async def _read(request: web.Request, json):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    stream = get_stream_from_request_params(request, db)
    
    # parse optional parameters
    params = {'start': None, 'end': None, 'max-rows': None, 'decimation-level': None}
    validate_query_parameters(params, request.query)
    # additional parameter checks
    if params['max-rows'] is not None and params['max-rows'] <= 0:
        raise web.HTTPBadRequest(reason="[max-rows] must be > 0")
    if params['decimation-level'] is not None and params['decimation-level'] <= 0:
        raise web.HTTPBadRequest(reason="[decimation-level] must be > 0")

    # --- Binary Streaming Handler ---
    resp = None

    async def stream_data(data: np.ndarray, layout, factor):
        nonlocal resp

        if resp is None:
            resp = web.StreamResponse(status=200,
                                      headers={'joule-layout': layout,
                                               'joule-decimation': str(factor)})
            resp.enable_chunked_encoding()
            await resp.prepare(request)
        await resp.write(data.tobytes())

    # --- JSON Handler ---

    data_blocks = []  # array of data segments
    data_segment = None
    decimation_factor = 1

    async def retrieve_data(data: np.ndarray, layout, factor):
        nonlocal data_blocks, data_segment, decimation_factor
        decimation_factor = factor
        ## CASE 1: Interval token by itself, no data
        if np.array_equal(data, pipes.interval_token(layout)):
            if data_segment is not None:
                data_blocks.append(data_segment.tolist())
                data_segment = None
        ## CASE 2: Data followed by an interval token
        elif len(data)>0 and np.array_equal([data[-1]],pipes.interval_token(layout)):
            # flatten the structured array
            data = np.c_[data['timestamp'][:, None], data['data']]
            # remove the interval token
            data = data[:-1,:]
            # CASE 2.1: Last segment of data from the interval (we already have part of it)
            if data_segment is not None:
                data_segment = np.vstack((data_segment, data))
                data_blocks.append(data_segment.tolist())
                data_segment = None
            # CASE 2.2: This chunk of data is the entire interval (no existing data from this interval)
            elif len(data)>0: 
                # only need to create a data block if there's actually data in this segment
                data_blocks.append(data.tolist())
        ## CASE 3: Just data- no interval token
        else:
            data = np.c_[data['timestamp'][:, None], data['data']]
            ## CASE 3.1: First segment of data from a new interval
            if data_segment is None:
                data_segment = data
            ## CASE 3.2: Subsequent segments of data from the same interval
            else:
                data_segment = np.vstack((data_segment, data))

    if json:
        callback = retrieve_data
    else:
        callback = stream_data

    # create an extraction task
    try:
        await data_store.extract(stream, params['start'], params['end'],
                                 callback=callback,
                                 max_rows=params['max-rows'],
                                 decimation_level=params['decimation-level'])
    except InsufficientDecimationError as e:
        raise web.HTTPBadRequest(reason="decimated data is not available: %s" % e)
    except DataError as e:
        raise web.HTTPBadRequest(reason=f"read error: {str(e)}")

    if json:
        # put the last data_segment on
        if data_segment is not None:
            data_blocks.append(data_segment.tolist())
        return web.json_response({"data": data_blocks, "decimation_factor": decimation_factor})
    else:
        if resp is None:
            return web.json_response(text="this stream has no data", status=400)
        return resp


async def _subscribe(request: web.Request, json: bool):
    db: Session = request.app[app_keys.db]
    supervisor: Supervisor = request.app[app_keys.supervisor]
    if json:
        raise web.HTTPBadRequest(reason="JSON subscription not implemented")
    stream = get_stream_from_request_params(request, db)
    pipe = pipes.LocalPipe(stream.layout)
    try:
        unsubscribe = supervisor.subscribe(stream, pipe)
    except SubscriptionError:
        raise web.HTTPBadRequest(reason="stream is not being produced")
    resp = web.StreamResponse(status=200,
                              headers={'joule-layout': stream.layout,
                                       'joule-decimation': '1'})
    resp.enable_chunked_encoding()

    try:
        await resp.prepare(request)
        while await pipe.not_empty():
            data = await pipe.read()
            pipe.consume(len(data))
            if len(data) > 0:
                await resp.write(data.tobytes())
            if pipe.end_of_interval:
                await resp.write(pipes.interval_token(stream.layout).tobytes())
    except ConnectionResetError:
        pass # ignore client disconnect
    finally:
        unsubscribe()
    return resp

async def intervals(request: web.Request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    stream = get_stream_from_request_params(request, db)
    params = {'start': None, 'end': None}
    validate_query_parameters(params, request.query)

    return web.json_response(await data_store.intervals(stream, params['start'], params['end']))


async def write(request: web.Request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    stream = get_stream_from_request_params(request, db)
    # spawn in inserter task
    stream.is_destination = True
    db.commit()

    pipe = pipes.InputPipe(name="inbound", stream=stream, reader=request.content)
    if 'merge-gap' in request.query:
        try:
            merge_gap = int(request.query['merge-gap'])
        except ValueError:
            raise web.HTTPBadRequest(reason="merge-gap must be an integer")
        
        if merge_gap < 0:
            raise web.HTTPBadRequest(reason="merge-gap must be >= 0")
    else:
        merge_gap = 0
    try:
        task = await data_store.spawn_inserter(stream, pipe, insert_period=0, merge_gap=merge_gap)
        await task
    # ConnectionError is raised when the client disconnects unexpectedly which can happen often (module stops, etc)
    except (DataError, ConnectionError) as e:
        stream.is_destination = False
        db.commit()
        raise web.HTTPBadRequest(reason=str(e))
    finally:
        stream.is_destination = False
        db.commit()
    return web.Response(text="ok")


async def remove(request: web.Request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    stream = get_stream_from_request_params(request, db)

    params = {'start': None, 'end': None}
    validate_query_parameters(params, request.query)

    await data_store.remove(stream, params['start'], params['end'])
    return web.Response(text="ok")


async def consolidate(request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    _validate_support(data_store)

    stream = get_stream_from_request_params(request, db)
    params = {'start': None, 'end': None}
    validate_query_parameters(params, request.query)
    
    # parse the max_gap parameter
    if 'max_gap' not in request.query:
        raise web.HTTPBadRequest(reason="specify max_gap as us integer")
    try:
        max_gap = int(request.query['max_gap'])
        if max_gap <= 0:
            raise ValueError()
    except ValueError:
        raise web.HTTPBadRequest(reason="max_gap must be postive integer")
    num_removed = await data_store.consolidate(stream, params['start'], params['end'], max_gap)
    return web.json_response(data={"num_consolidated": num_removed})


async def decimate(request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    stream = get_stream_from_request_params(request, db)
    _validate_support(data_store)
    await data_store.decimate(stream)
    return web.Response(text="ok")

async def drop_decimations(request):
    db: Session = request.app[app_keys.db]
    data_store: DataStore = request.app[app_keys.data_store]
    _validate_support(data_store)
    stream = get_stream_from_request_params(request, db)
    await data_store.drop_decimations(stream)
    return web.Response(text="ok")

### Helper Functions ###



def _validate_support(data_store: DataStore):
    # TimeScale supports decimation management (others may not)
    if not data_store.supports_decimation_management:
        raise web.HTTPBadRequest(reason="data store does not support decimation")


from aiohttp import web
from joule.constants import ApiErrorMessages
from joule.models import DataStream, folder
from sqlalchemy.orm import Session
import json

async def read_json(request: web.Request):
    if request.content_type != 'application/json':
        raise web.HTTPBadRequest(reason='content-type must be application/json')
    try:
        return await request.json()
    except json.JSONDecodeError:
        raise web.HTTPBadRequest(reason='invalid json')

def validate_query_parameters(params, query):
    # populate the params dictionary with query parameters if they exist
    # return the parsed json_filter parameter if it exists
    try:
        for param in params:
            if param in query:
                params[param] = int(query[param])
    except ValueError:
        raise web.HTTPBadRequest(reason=ApiErrorMessages.f_parameter_must_be_an_int.format(parameter=param))
    
    # make sure parameters make sense
    if ((params['start'] is not None and params['end'] is not None) and
            (params['start'] >= params['end'])):
        raise web.HTTPBadRequest(reason=ApiErrorMessages.start_must_be_before_end)

def get_stream_from_request_params(request: web.Request, db: Session) -> DataStream:
    if 'path' in request.query:
        stream = folder.find_stream_by_path(request.query['path'], db, stream_type=DataStream)
    elif 'id' in request.query:
        stream = db.get(DataStream,request.query["id"])
    else:
        raise web.HTTPBadRequest(reason="specify an id or a path")
    if stream is None:
        raise web.HTTPNotFound(reason="stream does not exist")
    return stream


import click
import asyncio

import joule.api
from joule.utilities import timestamp_to_human as ts2h
from joule.utilities import human_to_timestamp as h2ts
from joule import errors
from joule.api import get_node
from joule.cli.config import Config, pass_config


@click.command(name="copy")
@click.option('-s', "--start", help="timestamp or descriptive string")
@click.option('-e', "--end", help="timestamp or descriptive string")
@click.option('-a', '--action', help="action to take if events already exist in the destination",
              type=click.Choice(['ignore', 'replace', 'prompt']), default='prompt')
@click.option('-n', '--new', help="copy starts at the last timestamp of the destination", is_flag=True)
@click.option('-d', '--destination-node', help="node name")
@click.argument("source")
@click.argument("destination")
@pass_config
def cli_copy(config: Config, start, end, action, new, destination_node, source, destination):
    """Copy events to a different stream."""
    try:
        if destination_node is None:
            dest_node = config.node
        else:
            dest_node = get_node(destination_node)
    except errors.ApiError:
        raise click.ClickException(f"Invalid destination node [{destination_node}]")
    try:
        replace_action = asyncio.run(_wrap_action_for_existing_events(dest_node, destination, action, start, end))
        asyncio.run(_run(config.node, dest_node, start, end, new, replace_action, source, destination))
    except errors.ApiError as e:
        raise click.ClickException(str(e)) from e
    finally:
        asyncio.run(config.close_node())
        if destination_node is not None:  # different destination node
            asyncio.run(dest_node.close())
    click.echo("OK")


async def _wrap_action_for_existing_events(node, stream, action, start, end):
    try:
        result = await _action_for_existing_events(node, stream, action, start, end)
        return result
    finally:
        await node.close()


async def _action_for_existing_events(node, stream, action, start, end):
    if action == 'replace':
        return True
    elif action == 'ignore':
        return False
    elif action == 'prompt':
        if not await has_existing_events(node, stream, start, end):
            return False
        print("""
There are already events in this destination, select how you want to proceed:
[c]ancel: stop, do not copy anything
[i]gnore: ignore existing destination events, add source events. This may result in duplicate events
[r]eplace: remove all destination events, then add source events. This may result in data loss
Select an option (c,i or r): """)
        choice = click.getchar()
        if choice == 'c':
            raise click.ClickException("Action cancelled")
        elif choice == 'i':
            print("\t ignoring existing events, running copy anyway")
            return False
        elif choice == 'r':
            print("\t removing events in destination before running copy")
            return True
    else:
        raise click.ClickException("\t invalid option, cancelling copy")


async def has_existing_events(node, stream, start, end):
    try:
        count = await node.event_stream_count(stream, start=start, end=end, include_on_going_events=False)
        return count > 0
    except joule.errors.ApiError as e:
        if "does not exist" in str(e):
            return False
        else:
            raise e


async def _run(source_node, dest_node, start, end, new, replace, source, destination):
    # make sure the time bounds make sense
    if start is not None:
        try:
            start = h2ts(start)
        except ValueError:
            raise errors.ApiError("invalid start time: [%s]" % start)
    if end is not None:
        try:
            end = h2ts(end)
        except ValueError:
            raise errors.ApiError("invalid end time: [%s]" % end)
    if (start is not None) and (end is not None) and ((end - start) <= 0):
        raise click.ClickException(f"Error: start {ts2h(start)} " +
                                   f"must be before end f{ts2h(end)}")

    # create the destination stream if necessary
    # name = destination.split('/')[-1]
    # path = "/".join(destination.split('/')[:-1])
    source_stream = await source_node.event_stream_get(source)

    await dest_node.event_stream_get(destination, create=True,
                                     description=source_stream.description,
                                     event_fields=source_stream.event_fields,

                                     chunk_duration_us=source_stream.chunk_duration_us)
    if new:
        dest_info = await dest_node.event_stream_info(destination)
        start = dest_info.end
        if start is not None:
            print(f"Starting copy at {ts2h(dest_info.end)}")
    elif replace:
        print("Removing existing events from destination...", end="")
        await dest_node.event_stream_remove(destination, start, end)
        print("[OK]")

    stream_info = await source_node.event_stream_info(source)
    event_count = stream_info.event_count
    num_copied_events = 0
    with click.progressbar(length=event_count) as bar:
        async for events in source_node.event_stream_read(source, start=start, end=end,
                                                          include_on_going_events=False, block_size=10_000):
            await dest_node.event_stream_write(destination, events)
            num_copied_events += len(events)
            bar.update(len(events))
            start = events[-1].start_time + 1
        # bring bar up to 100%
        bar.update(event_count - num_copied_events)



import unittest
import configparser
import tempfile
import yarl
import testing.postgresql

from joule.services import load_config
from joule.models import config


class TestLoadConfigs(unittest.TestCase):

    def test_loads_default_config(self):
        my_config = load_config.run(verify=False)
        self.assertEqual(my_config.socket_directory, config.DEFAULT_CONFIG['Main']['SocketDirectory'])
        self.assertFalse(my_config.echo_module_logs)

    def test_customizes_config(self):
        parser = configparser.ConfigParser()
        parser.read_string("""
            [Main]
            IPAddress = 8.8.8.8
            Port = 8080
            Database = new_db
            InsertPeriod = 20
            EchoModuleLogs = yes
        """)
        my_config = load_config.run(custom_values=parser, verify=False)
        self.assertEqual(my_config.ip_address, "8.8.8.8")
        self.assertEqual(my_config.insert_period, 20)
        self.assertTrue(my_config.echo_module_logs)

    def test_verifies_directories_exist(self):
        postgresql = testing.postgresql.Postgresql()
        db_url = postgresql.url()
        with tempfile.TemporaryDirectory() as module_dir:
            with tempfile.TemporaryDirectory() as stream_dir:
                with tempfile.TemporaryDirectory() as socket_dir:

                    parser = configparser.ConfigParser()
                    parser.read_string("""
                                [Main]
                                ModuleDirectory=%s
                                StreamDirectory=%s
                                SocketDirectory=%s
                                Database=%s
                            """ % (module_dir, stream_dir, socket_dir, db_url[13:]))
                    my_config = load_config.run(custom_values=parser)
                    self.assertEqual(my_config.stream_directory, stream_dir)
                    self.assertEqual(my_config.module_directory, module_dir)
        postgresql.stop()

    def test_loads_proxies(self):
        parser = configparser.ConfigParser()
        parser.read_string("""
                    [Proxies]
                    site1=http://localhost:5000
                    site two=https://othersite.com
                """)
        my_config = load_config.run(custom_values=parser, verify=False)
        self.assertEqual(my_config.proxies[0].url, yarl.URL("http://localhost:5000"))
        self.assertEqual(my_config.proxies[0].uuid, 0)
        self.assertEqual(my_config.proxies[0].name, "site1")

        self.assertEqual(my_config.proxies[1].url, yarl.URL("https://othersite.com"))
        self.assertEqual(my_config.proxies[1].uuid, 1)
        self.assertEqual(my_config.proxies[1].name, "site two")



from click.testing import CliRunner
import os
import warnings
from tests.cli.fake_joule import FakeJoule, FakeJouleTestCase
from joule.cli import main

MODULE_LIST = os.path.join(os.path.dirname(__file__), 'modules.json')
warnings.simplefilter('always')


class TestModuleList(FakeJouleTestCase):

    def test_lists_modules(self):
        server = FakeJoule()
        with open(MODULE_LIST, 'r') as f:
            server.response = f.read()
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['module', 'list'])
        self.assertEqual(result.exit_code, 0)
        output = result.output
        # make sure modules are listed
        for name in ['Module%d' % x for x in range(1,5)]:
            self.assertIn(name, output)
        # check a few streams
        for stream in ['folder_1/stream_1_1', '/folder_2/stream_2_1']:
            self.assertIn(stream, output)
        self.stop_server()

    def test_when_server_returns_invalid_data(self):
        server = FakeJoule()
        server.response = "notjson"
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['module', 'list'])
        self.assertIn('Error', result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()

    def test_when_server_returns_error_code(self):
        server = FakeJoule()
        error_msg = "test error"
        error_code = 500
        server.response = error_msg
        server.http_code = error_code
        self.start_server(server)
        runner = CliRunner()
        result = runner.invoke(main, ['module', 'list'])
        self.assertIn('%d' % error_code, result.output)
        self.assertIn(error_msg, result.output)
        self.assertEqual(result.exit_code, 1)
        self.stop_server()



import asyncio

from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop
from aiohttp import web

import joule.controllers
from tests.controllers.helpers import create_db
from joule.models import DataStream, Annotation
from joule import utilities
import testing
import time
from joule import app_keys
from joule.constants import EndPoints

import testing.postgresql
psql_key = web.AppKey("psql", testing.postgresql.Postgresql)

class TestAnnotationController(AioHTTPTestCase):

    async def get_application(self):
        app = web.Application()
        app.add_routes(joule.controllers.routes)

        # this takes a while, adjust the expected coroutine execution time
        loop = asyncio.get_running_loop()
        loop.slow_callback_duration = 2.0

        db, app[psql_key] = create_db(["/top/leaf/stream1:float32[x, y, z]",
                                     "/top/middle/leaf/stream2:int16[val1, val2]"])
        self.stream1 = db.query(DataStream).filter_by(name="stream1").one_or_none()
        self.stream2 = db.query(DataStream).filter_by(name="stream2").one_or_none()

        # add 5 event annotations to stream1
        # add 5 interval annotations to stream2
        for x in range(5):
            # start: 0, 1000, 2000, 3000, 4000
            start = utilities.timestamp_to_datetime(x * 1000)
            end = utilities.timestamp_to_datetime((x * 1000) + 200)
            a1 = Annotation(title="stream1_%d" % x, start=start)
            a2 = Annotation(title="stream2_%d" % x, start=start, end=end)
            a1.stream = self.stream1
            a2.stream = self.stream2
            db.add(a1)
            db.add(a2)
        db.commit()
        app[app_keys.db] = db
        self.db = db
        return app

    async def test_annotation_list(self):
        # 1.) retreive all annotations
        resp = await self.client.request("GET", EndPoints.annotations,
                                         params=[("stream_id", str(self.stream1.id)),
                                                 ("stream_id", str(self.stream2.id))])
        all_annotations_json = await resp.json()
        self.assertEqual(len(all_annotations_json), 10)
        for annotation in all_annotations_json:
            if "stream1" in annotation["title"]:
                self.assertIsNone(annotation["end"])
            elif "stream2" in annotation["title"]:
                self.assertEqual(annotation["end"] - annotation["start"], 200)
            else:
                self.fail("invalid stream title")
        # 2.) retrieve a time bounded list
        resp = await self.client.request("GET", EndPoints.annotations,
                                         params={"stream_id": self.stream1.id,
                                                 "start": 500,
                                                 "end": 3500})
        annotations_json = await resp.json()
        # expect 1000, 2000, 3000 start timestamps
        self.assertEqual(len(annotations_json), 3)
        for annotation in annotations_json:
            self.assertIn("stream1", annotation["title"])
            self.assertGreater(annotation["start"], 500)
            self.assertLess(annotation["start"], 3500)

        # 3.) retrieve annotation by path
        resp = await self.client.request("GET", EndPoints.annotations,
                                         params=[("stream_id", self.stream1.id),
                                                 ("stream_path", "/top/middle/leaf/stream2")])
        annotations_json = await resp.json()
        self.assertEqual(annotations_json, all_annotations_json)

    async def test_annotation_info(self):
        resp = await self.client.request("GET", EndPoints.annotations_info,
                                         params={"stream_id": self.stream1.id})
        self.assertEqual(resp.status, 200)
        info = await resp.json()
        self.assertEqual(info["count"], 5)
        self.assertEqual(info["start"], 0)
        self.assertEqual(info["end"], 4000)

        # can get info by path
        resp = await self.client.request("GET", EndPoints.annotations_info,
                                            params={"stream_path": "/top/middle/leaf/stream2"})
        self.assertEqual(resp.status, 200)
        info = await resp.json()
        self.assertEqual(info["count"], 5)
        self.assertEqual(info["start"], 0)
        self.assertEqual(info["end"], 4000)

    async def test_annotation_create_by_stream_id(self):
        annotation_json = {
            "title": "new_annotation",
            "content": "content",
            "start": 900,
            "end": 1000,
            "stream_id": self.stream1.id
        }
        resp = await self.client.request("POST", EndPoints.annotation,
                                         json=annotation_json)
        values = await resp.json()
        new_id = values["id"]
        # make sure it was added
        new_annotation = self.db.get(Annotation,new_id)
        self.assertEqual(new_annotation.title, "new_annotation")
        self.assertEqual(new_annotation.stream_id, self.stream1.id)
        self.assertEqual(new_annotation.start, utilities.timestamp_to_datetime(900))
        self.assertEqual(new_annotation.end, utilities.timestamp_to_datetime(1000))

    async def test_annotation_create_by_stream_path(self):
        annotation_json = {
            "title": "new_annotation",
            "content": "content",
            "start": 900,
            "end": 1000,
            "stream_path": "/top/leaf/stream1"
        }
        resp = await self.client.request("POST", EndPoints.annotation,
                                         json=annotation_json)
        values = await resp.json()
        new_id = values["id"]
        # make sure it was added
        new_annotation = self.db.get(Annotation,new_id)
        self.assertEqual(new_annotation.title, "new_annotation")
        self.assertEqual(new_annotation.stream_id, self.stream1.id)
        self.assertEqual(new_annotation.start, utilities.timestamp_to_datetime(900))
        self.assertEqual(new_annotation.end, utilities.timestamp_to_datetime(1000))

    async def test_annotation_edit(self):
        old_annotation = self.db.query(Annotation).filter_by(stream_id=self.stream1.id).first()
        annotation_json = {
            "title": "updated",
            "content": "this is updated too",
            "start": utilities.datetime_to_timestamp(old_annotation.start) + 100,  # should not update
            "end": 1000,  # should not update
            "id": old_annotation.id
        }
        resp = await self.client.request("PUT", EndPoints.annotation,
                                         json=annotation_json)
        self.assertEqual(resp.status, 200)
        new_annotation = self.db.get(Annotation,old_annotation.id)
        self.assertEqual(new_annotation.title, "updated")
        self.assertEqual(new_annotation.content, "this is updated too")
        self.assertEqual(new_annotation.start, old_annotation.start)
        self.assertEqual(new_annotation.end, old_annotation.end)

    async def test_annotation_delete(self):
        old_annotation = self.db.query(Annotation).filter_by(stream_id=self.stream1.id).first()

        resp = await self.client.request("DELETE", EndPoints.annotation,
                                         params={"id": old_annotation.id})
        self.assertEqual(resp.status, 200)
        self.assertIsNone(self.db.query(Annotation).filter_by(id=old_annotation.id).one_or_none())

    async def test_annotation_delete_cascade(self):
        # make sure annotation are deleted when streams are deleted
        for stream in self.db.query(DataStream).all():
            self.db.delete(stream)
        self.db.commit()
        self.assertEqual(0, self.db.query(Annotation).count())

    async def test_annotation_delete_all(self):
        # delete a time range
        self.assertEqual(5, self.db.query(Annotation).
                         filter_by(stream_id=self.stream1.id).
                         count())
        resp = await self.client.request("DELETE", EndPoints.stream_annotations,
                                         params={"stream_id": self.stream1.id,
                                                 "start": 500,
                                                 "end": 3500})
        # expect 1000, 2000, 3000 to be deleted, just 2 left (0 and 4000)
        self.assertEqual(resp.status, 200)
        self.assertEqual(2, self.db.query(Annotation).
                         filter_by(stream_id=self.stream1.id).
                         count())
        annotations = self.db.query(Annotation).filter_by(stream_id=self.stream1.id).all()
        for a in annotations:
            ts = utilities.datetime_to_timestamp(a.start)
            if (ts > 500) and (ts < 3500):
                self.fail("this annotation should be deleted")

        # delete by stream_id
        resp = await self.client.request("DELETE", EndPoints.stream_annotations,
                                         params={"stream_id": self.stream1.id})
        self.assertEqual(resp.status, 200)
        self.assertEqual(0, self.db.query(Annotation).
                         filter_by(stream_id=self.stream1.id).
                         count())
        # delete by path
        self.assertEqual(5, self.db.query(Annotation).
                         filter_by(stream_id=self.stream2.id).
                         count())
        resp = await self.client.request("DELETE", EndPoints.stream_annotations,
                                         params={"stream_path": "/top/middle/leaf/stream2"})
        self.assertEqual(resp.status, 200)
        self.assertEqual(0, self.db.query(Annotation).
                         filter_by(stream_id=self.stream2.id).
                         count())

        # make sure the info is updated
        resp = await self.client.request("GET", EndPoints.annotations_info,
                                         params={"stream_id": self.stream1.id})
        info = await resp.json()
        self.assertEqual(info["count"], 0)
        self.assertIsNone(info["start"])
        self.assertIsNone(info["end"])


from copy import copy


class MockSession:

    def __init__(self, multiple_calls=False):
        self.path = None,
        self.response_data = None
        self.request_data = None
        self.method = None
        self.methods = []
        self.paths = []
        self.multiple_calls = multiple_calls

    async def put(self, path, json):
        self.path = path
        self.paths.append(path)
        self._save_request_data(json)
        self.method = 'PUT'
        self.methods.append(self.method)
        return self._response_data()

    async def delete(self, path, params):
        self.path = path
        self.paths.append(path)
        self._save_request_data(params)
        self.method = 'DELETE'
        self.methods.append(self.method)
        return self._response_data()

    async def get(self, path, params=None):
        self.path = path
        self.paths.append(path)
        self._save_request_data(params)
        self.method = 'GET'
        self.methods.append(self.method)
        return self._response_data()

    async def post(self, path, json):
        self.path = path
        self.paths.append(path)
        self._save_request_data(json)
        self.method = 'POST'
        self.methods.append(self.method)
        return self._response_data()

    def _save_request_data(self, data):
        #print(f"data id: {id(data)}")
        if self.multiple_calls:
            if self.request_data is None:
                self.request_data = []
            self.request_data.append(copy(data))
        else:
            self.request_data = data

    def _response_data(self):
        if self.multiple_calls:
            response = self.response_data.pop(0)
        else:
            response = self.response_data
        if issubclass(type(response), Exception):
            raise response
        return response



from joule.models.pipes import LocalPipe
import asyncio
import numpy as np
import argparse
import tempfile
import os
import time

from tests import helpers
from joule.client.builtins.file_reader import FileReader

WIDTH = 3


class TestFileReader(helpers.AsyncTestCase):

    def test_reads_timestamped_values(self):
        data = helpers.create_data("float32_8", length=3)
        (data_file, path) = tempfile.mkstemp()
        with open(data_file, 'w') as f:
            for row in data:
                f.write("%d %s\n" % (row['timestamp'], ' '.join(repr(x) for x in row['data'])))
        my_reader = FileReader()
        ''
        pipe = LocalPipe("float32_8", name="output")
        args = argparse.Namespace(file=path, delimiter=" ",
                                  timestamp=False)
        asyncio.run(my_reader.run(args, pipe))
        # check the results
        result = pipe.read_nowait()
        np.testing.assert_array_almost_equal(data['timestamp'], result['timestamp'])
        np.testing.assert_array_almost_equal(data['data'], result['data'])
        os.remove(path)

    def test_timestamps_raw_values(self):
        data = helpers.create_data("float32_8", length=3)
        (data_file, path) = tempfile.mkstemp()
        with open(data_file, 'w') as f:
            for row in data:
                f.write("%s\n" % ' '.join(repr(x) for x in row['data']))
        my_reader = FileReader()
        ''
        pipe = LocalPipe("float32_8", name="output")
        args = argparse.Namespace(file=path, delimiter=" ",
                                  timestamp=True)
        asyncio.run(my_reader.run(args, pipe))
        # check the results
        result = pipe.read_nowait()
        # timestamps should be close to now
        actual = np.average(result['timestamp'])
        expected = int(time.time() * 1e6)
        # should be within 1 second
        np.testing.assert_almost_equal(actual / 1e6, expected / 1e6, decimal=0)
        np.testing.assert_array_almost_equal(data['data'], result['data'])
        os.remove(path)

    def test_stops_on_request(self):
        data = helpers.create_data("float32_8")
        (data_file, path) = tempfile.mkstemp()
        with open(data_file, 'w') as f:
            for row in data:
                f.write("%d %s\n" % (row['timestamp'], ' '.join(repr(x) for x in row['data'])))
        my_reader = FileReader()
        ''
        pipe = LocalPipe("float32_8", name="output")
        args = argparse.Namespace(file=path, delimiter=" ",
                                  timestamp=False)
        my_reader.stop()
        asyncio.run(my_reader.run(args, pipe))
        # check the results
        result = pipe.read_nowait()

        # output should be shorter than the input
        self.assertLess(len(result), len(data))
        os.remove(path)



from typing import Optional, Dict
import json


class Event:
    def __init__(self,
                 start_time: int,
                 end_time: Optional[int] = None,
                 content: Optional[Dict] = None,
                 id: Optional[int] = None,
                 event_stream_id: Optional[int] = None,
                 node_uuid: Optional[str] = None):
        self.id = id
        self.start_time = int(start_time)
        self.end_time = int(end_time)
        # The two fields below are local to the API client and are needed
        # to distinguish events between different streams and nodes so that
        # write operations can be performed correctly- if an event has an ID
        # an update will be attempted but it is important that the ID corresponds
        # to the target table, if it has been copied from another stream this
        # could create an error where the update effectively erases an event
        # in the target table
        self.event_stream_id = event_stream_id
        self.node_uuid = node_uuid
        if content is None:
            self.content = {}
        else:
            self.content = content

    def to_json(self, destination_node_uuid: str):
        # remove the ID fields if the event is not destined for the same node
        if self.node_uuid != destination_node_uuid:
            _id_value = None
            _event_stream_id_value = None
        else:
            _id_value = self.id
            _event_stream_id_value = self.event_stream_id
        return {
            'id': _id_value,
            'event_stream_id': _event_stream_id_value,
            'start_time': self.start_time,
            'end_time': self.end_time,
            'content': self.content
        }

    def __repr__(self):
        content_str = json.dumps(self.content)
        if len(content_str) > 20:
            content_str = content_str[:20] + " ...}"
        return "<joule.api.Event start_time: %r, end_time: %r, content: %s>" % (
            self.start_time, self.end_time, content_str)

    def __eq__(self, other):
        if self.start_time != other.start_time:
            return False
        if self.end_time != other.end_time:
            return False
        if json.dumps(self.content) != json.dumps(other.content):
            return False
        return True


def from_json(json: dict, node_uuid: str=""):
    return Event(start_time=json['start_time'],
                 end_time=json['end_time'],
                 content=json['content'],
                 id=json['id'],
                 event_stream_id=json['event_stream_id'],
                 node_uuid=node_uuid)
